import os
import shutil
from functools import wraps
from os import path

import pytest

import autofit as af
from autoconf import conf
from autofit.mock import mock
from autofit.non_linear.nest import multi_nest as mn

directory = path.dirname(path.realpath(__file__))
pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


@pytest.fixture(name="multi_nest_summary_path")
def test_multi_nest_summary():
    multi_nest_summary_path = path.join(conf.instance.output_path, "non_linear", "multinest", "summary")

    if path.exists(multi_nest_summary_path):
        shutil.rmtree(multi_nest_summary_path)

    os.mkdir(multi_nest_summary_path)

    return multi_nest_summary_path


@pytest.fixture(name="multi_nest_samples_path")
def test_multi_nest_samples():
    multi_nest_samples_path = path.join(conf.instance.output_path, "non_linear", "multinest", "samples")

    if path.exists(multi_nest_samples_path):
        shutil.rmtree(multi_nest_samples_path)

    os.mkdir(multi_nest_samples_path)

    return multi_nest_samples_path


@pytest.fixture(name="multi_nest_resume_path")
def test_multi_nest_resume():
    multi_nest_resume_path = path.join(conf.instance.output_path, "non_linear", "multinest", "resume")

    if path.exists(multi_nest_resume_path):
        shutil.rmtree(multi_nest_resume_path)

    os.mkdir(multi_nest_resume_path)

    return multi_nest_resume_path


def create_path(func):
    @wraps(func)
    def wrapper(file_path):
        if not path.exists(file_path):
            os.makedirs(file_path)
        return func(file_path)

    return wrapper


@create_path
def create_summary_4_parameters(file_path):
    summary = open(path.join(file_path, "multinestsummary.txt"), "w")
    summary.write(
        "    0.100000000000000000E+01   -0.200000000000000000E+01    0.300000000000000000E+01"
        "    0.400000000000000000E+01   -0.500000000000000000E+01    0.600000000000000000E+01"
        "    0.700000000000000000E+01    0.800000000000000000E+01"
        "    0.900000000000000000E+01   -1.000000000000000000E+01   -1.100000000000000000E+01"
        "    1.200000000000000000E+01    1.300000000000000000E+01   -1.400000000000000000E+01"
        "   -1.500000000000000000E+01    1.600000000000000000E+01"
        "    0.020000000000000000E+00    0.999999990000000000E+07"
        "    0.020000000000000000E+00    0.999999990000000000E+07\n"
    )
    summary.write(
        "    0.100000000000000000E+01   -0.200000000000000000E+01    0.300000000000000000E+01"
        "    0.400000000000000000E+01   -0.500000000000000000E+01    0.600000000000000000E+01"
        "    0.700000000000000000E+01    0.800000000000000000E+01"
        "    0.900000000000000000E+01   -1.000000000000000000E+01   -1.100000000000000000E+01"
        "    1.200000000000000000E+01    1.300000000000000000E+01   -1.400000000000000000E+01"
        "   -1.500000000000000000E+01    1.600000000000000000E+01"
        "    0.020000000000000000E+00    0.999999990000000000E+07"
    )
    summary.close()


@create_path
def create_weighted_samples_4_parameters(file_path):
    with open(path.join(file_path, "multinest.txt"), "w+") as weighted_samples:
        weighted_samples.write(
            "    0.020000000000000000E+00    0.999999990000000000E+07    0.110000000000000000E+01    "
            "0.210000000000000000E+01    0.310000000000000000E+01    0.410000000000000000E+01\n"
            "    0.020000000000000000E+00    0.999999990000000000E+07    0.090000000000000000E+01    "
            "0.190000000000000000E+01    0.290000000000000000E+01    0.390000000000000000E+01\n"
            "    0.010000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    "
            "0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01\n"
            "    0.050000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    "
            "0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01\n"
            "    0.100000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    "
            "0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01\n"
            "    0.100000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    "
            "0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01\n"
            "    0.100000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    "
            "0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01\n"
            "    0.100000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    "
            "0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01\n"
            "    0.200000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    "
            "0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01\n"
            "    0.300000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    "
            "0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01"
        )


@create_path
def create_resume(file_path):
    with open(path.join(file_path, "multinestresume.dat"), "w+") as resume:
        resume.write(
            " F\n"
            "        3000       12345           1          50\n"
            "    0.502352236277967168E+05    0.502900436569068333E+05\n"
            " T\n"
            "   0\n"
            " T F     0          50\n"
            "    0.648698272260014622E-26    0.502352236277967168E+05    0.502900436569068333E+05\n"
        )


class TestMulitNest:
    def test__loads_from_config_file_if_not_input(self):
        multi_nest = af.MultiNest(
            prior_passer=af.PriorPasser(sigma=2.0, use_errors=False, use_widths=False),
            n_live_points=40,
            sampling_efficiency=0.5,
            const_efficiency_mode=True,
            evidence_tolerance=0.4,
            importance_nested_sampling=False,
            multimodal=False,
            n_iter_before_update=90,
            null_log_evidence=-1.0e80,
            max_modes=50,
            mode_tolerance=-1e88,
            seed=0,
            verbose=True,
            resume=False,
            context=1,
            write_output=False,
            log_zero=-1e90,
            max_iter=1,
            init_MPI=True,
        )

        assert multi_nest.prior_passer.sigma == 2.0
        assert multi_nest.prior_passer.use_errors is False
        assert multi_nest.prior_passer.use_widths is False
        assert multi_nest.config_dict["n_live_points"] == 40
        assert multi_nest.config_dict["sampling_efficiency"] == 0.5

        multi_nest = af.MultiNest()

        assert multi_nest.prior_passer.sigma == 3.0
        assert multi_nest.prior_passer.use_errors is True
        assert multi_nest.prior_passer.use_widths is True
        assert multi_nest.config_dict["n_live_points"] == 50
        assert multi_nest.config_dict["sampling_efficiency"] == 0.6

        model = af.ModelMapper(mock_class_1=mock.MockClassx4)

        fitness = af.MultiNest.Fitness(
            analysis=None,
            model=model,
            samples_from_model=multi_nest.samples_via_sampler_from_model,
            stagger_resampling_likelihood=False,
            paths=None
        )

        assert fitness.model == model

    def test__read_quantities_from_weighted_samples_file(self, multi_nest_samples_path):
        multi_nest = af.MultiNest()
        multi_nest.paths = af.DirectoryPaths(path_prefix=path.join("non_linear", "multinest"))

        create_weighted_samples_4_parameters(file_path=multi_nest.paths.path)

        parameters = mn.parameters_from_file_weighted_samples(
            file_weighted_samples=path.join(multi_nest.paths.path, "multinest.txt"),
            prior_count=4,
        )

        assert parameters == [
            [1.1, 2.1, 3.1, 4.1],
            [0.9, 1.9, 2.9, 3.9],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
        ]

        log_likelihoods = mn.log_likelihoods_from_file_weighted_samples(
            file_weighted_samples=path.join(multi_nest.paths.path, "multinest.txt")
        )

        value = -0.5 * 9999999.9

        assert log_likelihoods == 10 * [value]

        weights = mn.weights_from_file_weighted_samples(
            file_weighted_samples=path.join(multi_nest.paths.path, "multinest.txt")
        )

        assert weights == [0.02, 0.02, 0.01, 0.05, 0.1, 0.1, 0.1, 0.1, 0.2, 0.3]

    def test__read_total_samples_from_file_resume(self, multi_nest_resume_path):
        multi_nest = af.MultiNest()

        create_resume(file_path=multi_nest.paths.path)

        total_samples = mn.total_samples_from_file_resume(
            file_resume=path.join(multi_nest.paths.path, "multinestresume.dat")
        )

        assert total_samples == 12345

    def test__log_evidence_from_file_summary(self, multi_nest_summary_path):
        multi_nest = af.MultiNest()
        multi_nest.paths = af.DirectoryPaths(path_prefix=path.join("non_linear", "multinest"))

        create_summary_4_parameters(file_path=multi_nest.paths.samples_path)

        log_evidence = mn.log_evidence_from_file_summary(
            file_summary=path.join(multi_nest.paths.samples_path, "multinestsummary.txt"),
            prior_count=4,
        )

        assert log_evidence == 0.02

    def test__samples_from_model(
            self, multi_nest_samples_path, multi_nest_resume_path, multi_nest_summary_path
    ):
        multi_nest = af.MultiNest()
        multi_nest.paths = af.DirectoryPaths(path_prefix=path.join("non_linear", "multinest"))

        create_weighted_samples_4_parameters(file_path=multi_nest.paths.samples_path)
        create_resume(file_path=multi_nest.paths.samples_path)
        create_summary_4_parameters(file_path=multi_nest.paths.samples_path)

        model = af.ModelMapper(mock_class=mock.MockClassx4)
        model.mock_class.two = af.LogUniformPrior(lower_limit=1e-8, upper_limit=10.0)

        samples = multi_nest.samples_via_sampler_from_model(model=model)

        assert samples.parameters == [
            [1.1, 2.1, 3.1, 4.1],
            [0.9, 1.9, 2.9, 3.9],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
        ]

        value = -0.5 * 9999999.9

        assert samples.log_likelihoods == 10 * [value]
        assert samples.log_priors == pytest.approx(
            [0.243902, 0.256410, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25], 1.0e-4
        )
        assert samples.weights == [0.02, 0.02, 0.01, 0.05, 0.1, 0.1, 0.1, 0.1, 0.2, 0.3]
        assert samples.total_samples == 12345
        assert samples.log_evidence == 0.02
        assert samples.number_live_points == 50



import os
from os import path
import shutil

import pytest

import autofit as af
from autofit.text import formatter as frm
from autoconf import conf

directory = path.dirname(path.realpath(__file__))

text_path = path.join("{}".format(path.dirname(path.realpath(__file__))), "files", "text")


@pytest.fixture(scope="session", autouse=True)
def do_something():
    conf.instance.push(path.join("{}".format(directory), "files", "config", "text"))


def test__value_result_string():

    str0 = frm.value_result_string_from(parameter_name="param0", value=2.0)
    assert str0 == "2.00"

    str0 = frm.value_result_string_from(parameter_name="param11", value=3.00)

    assert str0 == "3.0000"

    str0 = frm.value_result_string_from(
        parameter_name="param0", value=2.0, values_at_sigma=(1.5, 2.5)
    )
    assert str0 == "2.00 (1.50, 2.50)"

    str0 = frm.value_result_string_from(
        parameter_name="param0", value=2.0, unit="arcsec"
    )

    assert str0 == "2.00 arcsec"


def test__parameter_result_latex():

    str0 = frm.parameter_result_latex_from(parameter_name="param0", value=2.0)
    assert str0 == r"param0 = 2.00 & "

    str0 = frm.parameter_result_latex_from(
        parameter_name="param0", value=2.0, errors=(0.1, 0.2)
    )
    assert str0 == r"param0 = 2.00^{+0.20}_{-0.10} & "

    str0 = frm.parameter_result_latex_from(
        parameter_name="param0", value=3.00, subscript="a", errors=(0.1, 0.2)
    )

    assert str0 == r"param0_{\mathrm{a}} = 3.00^{+0.20}_{-0.10} & "

    str0 = frm.parameter_result_latex_from(
        parameter_name="param0", value=3.00, subscript="a", name_to_label=True
    )

    assert str0 == r"p0_{\mathrm{a}} = 3.00 & "

    str0 = frm.parameter_result_latex_from(
        parameter_name="param0", value=3.00, subscript="a", unit="kg"
    )

    assert str0 == r"param0_{\mathrm{a}} = 3.00 kg & "


def test__output_list_of_strings_to_file():
    if path.exists(text_path):
        shutil.rmtree(text_path)

    os.mkdir(text_path)

    results = ["hi\n", "hello"]
    frm.output_list_of_strings_to_file(
        file=text_path + "model.results", list_of_strings=results
    )

    file = open(text_path + "model.results", "r")

    assert file.readlines() == ["hi\n", "hello"]


def test_string():
    assert frm.format_string_for_parameter_name("radius") == "{:.2f}"
    assert frm.format_string_for_parameter_name("mass") == "{:.2f}"


def test_substring():
    assert frm.format_string_for_parameter_name("einstein_radius") == "{:.2f}"
    assert frm.format_string_for_parameter_name("mass_value_something") == "{:.2f}"



import os
import shutil
from importlib import reload
from os import path

import pytest

import autofit as af

link_dir = path.join("{}".format(path.dirname(path.realpath(__file__))), "files")

temp_folder_path = path.join(link_dir, "linked_folder")


@pytest.fixture(
    autouse=True
)
def make_files_directory():
    try:
        os.mkdir(link_dir)
    except FileExistsError:
        pass
    yield
    shutil.rmtree(link_dir, ignore_errors=True)


def delete_trees(*paths):
    for path in paths:
        try:
            shutil.rmtree(path)
        except OSError:
            os.remove(path)


class TestCase:
    def test_create_dir(self):
        assert path.exists(af.link.autolens_dir)

    def test_environment_model(self):
        symdir = path.join("~", ".symdir")
        os.environ["SYMDIR"] = symdir
        reload(af.link)
        assert ".symdir" in af.link.autolens_dir
        assert path.exists(af.link.autolens_dir)

    def test_consistent_dir(self):
        directory = af.link.path_for(path.join("a", "random", "directory"))
        assert af.link.autolens_dir in directory
        assert directory == af.link.path_for(path.join("a", "random", "directory"))
        assert directory != af.link.path_for(path.join("b", "random", "directory"))
        assert af.link.path_for(path.join("linked_file")) != af.link.path_for(
            path.join("linked_folder")
        )

        link_1 = af.link.path_for(
            path.join(
                "",
                "abcdefghijklmnopqrstuv",
                "a",
                "long",
                "directory",
                "that",
                "has",
                "a",
                "difference",
                "in",
                "the",
                "middle",
                "of",
                "the",
                "path" "",
                "abcdefghijklmnopqrstuv",
            )
        )

        link_2 = af.link.path_for(
            path.join(
                "",
                "abcdefghijklmnopqrstuv",
                "a",
                "long",
                "directory",
                "that",
                "is",
                "different",
                "in",
                "the",
                "middle",
                "of",
                "the",
                "path" "",
                "abcdefghijklmnopqrstuv",
            )
        )

        assert link_1 != link_2

    def test_make_linked_folder(self):

        if path.exists(temp_folder_path):
            os.remove(temp_folder_path)

        linked_path = af.link.make_linked_folder(temp_folder_path)
        assert af.link.autolens_dir in linked_path
        assert path.exists(linked_path)
        assert path.exists(temp_folder_path)
        delete_trees(linked_path, temp_folder_path)

    def test_longer_path(self):
        long_folder_path = path.join("tmp", "folder", "path")
        with pytest.raises(FileNotFoundError):
            af.link.make_linked_folder(long_folder_path)

    def test_clean_source(self):
        linked_path = af.link.make_linked_folder(temp_folder_path)
        temp_file_path = path.join("{}".format(linked_path), "{}".format("temp"))
        open(temp_file_path, "a").close()
        assert path.exists(temp_file_path)

        delete_trees(temp_folder_path)
        assert not path.exists(temp_folder_path)
        af.link.make_linked_folder(temp_folder_path)
        assert not path.exists(temp_file_path)

    def test_not_clean_source(self):
        linked_path = af.link.make_linked_folder(temp_folder_path)
        temp_file_path = path.join("{}".format(linked_path), "{}".format("temp"))
        open(temp_file_path, "a").close()
        assert path.exists(temp_file_path)

        af.link.make_linked_folder(temp_folder_path)
        assert path.exists(temp_file_path)



from os import path

import pytest

import autofit as af
from autofit.mock.mock import MockClassx2
from autofit.non_linear.samples import Sample
from autofit.text import samples_text

text_path = path.join("{}".format(path.dirname(path.realpath(__file__))), "files", "samples")


@pytest.fixture(name="model")
def make_model():
    return af.ModelMapper(mock_class=MockClassx2)


@pytest.fixture(name="samples")
def make_samples(model):
    parameters = [[1.0, 2.0], [1.2, 2.2]]

    log_likelihoods = [1.0, 0.0]

    return af.PDFSamples(
        model=model,
        samples=Sample.from_lists(
            parameters=parameters,
            log_likelihoods=log_likelihoods,
            log_priors=[0.0, 0.0],
            weights=log_likelihoods,
            model=model
        )
    )


def test__summary(samples):
    results_at_sigma = samples_text.summary(samples=samples, sigma=3.0)

    assert "one       1.00 (1.00, 1.20)" in results_at_sigma
    assert "two       2.00 (2.00, 2.20)" in results_at_sigma


def test__latex(samples):
    latex_results_at_sigma = samples_text.latex(samples=samples, sigma=3.0)

    assert r"one_label_{\mathrm{o}} = 1.00^{+0.20}_{-0.00} & " in latex_results_at_sigma
    assert r"two_label_{\mathrm{t}} = 2.00^{+0.20}_{-0.00}" in latex_results_at_sigma



import numpy as np
import pytest
from scipy import stats, integrate

import numpy as np

import autofit.graphical.messages.normal
from autofit import graphical as graph

def test_message_norm():
    messages = [
        tuple(
            map(graph.NormalMessage, 
                [0.5, 0.1], [0.2, 0.3])),
        tuple(
            map(graph.NormalMessage, 
                [0.5, 0.1, -0.5], [0.2, 0.3, 1.3])),
        tuple(
            map(graph.GammaMessage, 
                [0.5, 1.1], [0.2, 1.3])),
        tuple(
            map(graph.GammaMessage, 
                [0.5, 1.1, 2], [0.2, 1.3, 1])),
    ]
    for ms in messages:
        m1, *m2s = ms
        A = np.exp(m1.log_normalisation(*m2s))

        # Calculate normalisation numerically
        i12, ierr = integrate.quad(
            lambda x: np.exp(sum(m.logpdf(x) for m in ms)), 
            *m1._support[0])

        # verify within tolerance
        assert np.abs(A - i12) < ierr < 1e-6

def test_numerical_gradient_hessians():
    N = graph.NormalMessage
    test_cases = [
        (N, 1., 0.5, 0.3),
        (N, 1., 0.5, [0.3, 2.1]),
        (N, [0.1, 1., 2.], [2., 0.5, 3.], [0.1, 0.2, 0.3]),
        (N, [0.1, 1., 2.], [2., 0.5, 3.], [[0.1, 0.2, 0.3], [2., 1., -1]]),
    ]
    for M, m, s, x in test_cases:
        message = M(m, s)
        res = message.logpdf_gradient_hessian(x)
        nres = message.numerical_logpdf_gradient_hessian(x)

        for a, n in zip(res, nres):
            assert np.linalg.norm(a - n) == pytest.approx(0, abs=1e-2)


def test_meanfield_gradients():
    n1, n2, n3 = 2, 3, 5
    p1, p2, p3 = [graph.Plate() for i in range(3)]

    v1 = graph.Variable('v1', p1, p2)
    v2 = graph.Variable('v2', p2, p3)
    v3 = graph.Variable('v3', p3, p1)

    mean_field = graph.MeanField({
        v1: graph.NormalMessage(
            np.random.randn(n1, n2),
            np.random.exponential(size=(n1, n2))),
        v2: graph.NormalMessage(
            np.random.randn(n2, n3),
            np.random.exponential(size=(n2, n3))),
        v3: graph.NormalMessage(
            np.random.randn(n3, n1),
            np.random.exponential(size=(n3, n1)))})

    values = mean_field.sample()
    l0 = mean_field(values, axis=None)
    logl = mean_field(values, axis=False)
    assert logl.sum() == pytest.approx(l0, abs=1e-5)
    logl = mean_field(values, axis=1)
    assert logl.sum() == pytest.approx(l0, abs=1e-5)
    logl = mean_field(values, axis=2)
    assert logl.sum() == pytest.approx(l0, abs=1e-5)
    logl = mean_field(values, axis=(0, 2))
    assert logl.sum() == pytest.approx(l0, abs=1e-5)

    njac0 = mean_field._numerical_func_jacobian(
        values, axis=None, _eps=1e-8)[1]
    njac1 = mean_field._numerical_func_jacobian(
        values, axis=1, _eps=1e-8)[1]
    njac2 = mean_field._numerical_func_jacobian(
        values, axis=(0, 1), _eps=1e-8)[1]
    njac = mean_field._numerical_func_jacobian_hessian(
        values, axis=False, _eps=1e-8)[1]
    grad = mean_field.logpdf_gradient(values, axis=False)[1]
    for v in grad:
        norm = np.linalg.norm(grad[v] - njac[v].sum((0, 1, 2)))
        assert norm == pytest.approx(0, abs=1e-2)
        norm = np.linalg.norm(grad[v] - njac0[v])
        assert norm == pytest.approx(0, abs=1e-2)
        norm = np.linalg.norm(grad[v] - njac1[v].sum((0, 1)))
        assert norm == pytest.approx(0, abs=1e-2)
        norm = np.linalg.norm(grad[v] - njac2[v].sum(0))
        assert norm == pytest.approx(0, abs=1e-2)


import numpy as np
from contextlib import contextmanager
import sys, os
import json

@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout

def numpy_array_to_json(
    array: np.ndarray, file_path: str, overwrite: bool = False
):
    """
    Write a NumPy array to a json file.

    Parameters
    ----------
    array : np.ndarray
        The array that is written to json.
    file_path : str
        The full path of the file that is output, including the file name and `.json` extension.
    overwrite : bool
        If `True` and a file already exists with the input file_path the .json file is overwritten. If 
        `False`, an error will be raised.

    Returns
    -------
    None

    Examples
    --------
    array_2d = np.ones((5,5))
    numpy_array_to_json(array_2d=array_2d, file_path='/path/to/file/filename.json', overwrite=True)
    """

    file_dir = os.path.split(file_path)[0]

    if not os.path.exists(file_dir):
        os.makedirs(file_dir)

    if overwrite and os.path.exists(file_path):
        os.remove(file_path)

    with open(file_path, "w+") as f:
        json.dump(array.tolist(), f)

def numpy_array_from_json(file_path: str):
    """
    Read a 1D NumPy array from a .json file.

    After loading the NumPy array, the array is flipped upside-down using np.flipud. This is so that the structures
    appear the same orientation as .json files loaded in DS9.

    Parameters
    ----------
    file_path : str
        The full path of the file that is loaded, including the file name and ``.json`` extension.

    Returns
    -------
    ndarray
        The NumPy array that is loaded from the .json file.

    Examples
    --------
    array_2d = numpy_array_from_json(file_path='/path/to/file/filename.json')
    """
    with open(file_path, "r") as f:
        return np.asarray(json.load(f))



import pytest

import autofit as af
from autofit import database as m
from autofit.mock import mock
from autofit.non_linear.samples import Sample


@pytest.fixture(
    name="save_samples",
    autouse=True
)
def save_samples(
        paths,
        sample
):
    samples = af.OptimizerSamples(
        model=af.Model(
            mock.Gaussian
        ),
        samples=[sample]
    )
    paths.save_samples(
        samples
    )


@pytest.fixture(
    name="sample"
)
def make_sample():
    return Sample(
        log_likelihood=1.0,
        log_prior=1.0,
        weights=0.5,
        centre=1.0,
        intensity=2.0,
        sigma=3.0
    )


def test_serialise_sample(sample):
    assert isinstance(
        sample.kwargs,
        dict
    )

    sample = m.Object.from_object(
        sample
    )()
    assert "centre" in sample.kwargs


def test_load_samples(
        paths
):
    samples = paths._load_samples()

    assert samples.model.cls is mock.Gaussian

    sample, = samples.samples
    assert sample.weights == 0.5



import autofit as af
from autofit.mock.mock import MockClassx4


class MockFitness:
    def figure_of_merit_from_parameters(self, parameters):
        return 1.0


class TestInitializePrior:
    def test__prior__initial_samples_sample_priors(self):

        model = af.PriorModel(MockClassx4)
        model.one = af.UniformPrior(lower_limit=0.099, upper_limit=0.101)
        model.two = af.UniformPrior(lower_limit=0.199, upper_limit=0.201)
        model.three = af.UniformPrior(lower_limit=0.299, upper_limit=0.301)
        model.four = af.UniformPrior(lower_limit=0.399, upper_limit=0.401)

        initializer = af.InitializerPrior()

        initial_unit_parameters, initial_parameters, initial_figures_of_merit = initializer.initial_samples_from_model(
            total_points=2, model=model, fitness_function=MockFitness()
        )

        assert 0.0 < initial_unit_parameters[0][0] < 1.0
        assert 0.0 < initial_unit_parameters[1][0] < 1.0
        assert 0.0 < initial_unit_parameters[0][1] < 1.0
        assert 0.0 < initial_unit_parameters[1][1] < 1.0
        assert 0.0 < initial_unit_parameters[0][2] < 1.0
        assert 0.0 < initial_unit_parameters[1][2] < 1.0
        assert 0.0 < initial_unit_parameters[0][3] < 1.0
        assert 0.0 < initial_unit_parameters[1][3] < 1.0

        assert 0.099 < initial_parameters[0][0] < 0.101
        assert 0.099 < initial_parameters[1][0] < 0.101
        assert 0.199 < initial_parameters[0][1] < 0.201
        assert 0.199 < initial_parameters[1][1] < 0.201
        assert 0.299 < initial_parameters[0][2] < 0.301
        assert 0.299 < initial_parameters[1][2] < 0.301
        assert 0.399 < initial_parameters[0][3] < 0.401
        assert 0.399 < initial_parameters[1][3] < 0.401

        assert initial_figures_of_merit == [1.0, 1.0]

    def test__initial_samples_in_test_model(self):

        model = af.PriorModel(MockClassx4)
        model.one = af.UniformPrior(lower_limit=0.099, upper_limit=0.101)
        model.two = af.UniformPrior(lower_limit=0.199, upper_limit=0.201)
        model.three = af.UniformPrior(lower_limit=0.299, upper_limit=0.301)
        model.four = af.UniformPrior(lower_limit=0.399, upper_limit=0.401)

        initializer = af.InitializerPrior()

        initial_unit_parameters, initial_parameters, initial_figures_of_merit = initializer.initial_samples_in_test_mode(
            total_points=2, model=model,
        )

        assert 0.0 < initial_unit_parameters[0][0] < 1.0
        assert 0.0 < initial_unit_parameters[1][0] < 1.0
        assert 0.0 < initial_unit_parameters[0][1] < 1.0
        assert 0.0 < initial_unit_parameters[1][1] < 1.0
        assert 0.0 < initial_unit_parameters[0][2] < 1.0
        assert 0.0 < initial_unit_parameters[1][2] < 1.0
        assert 0.0 < initial_unit_parameters[0][3] < 1.0
        assert 0.0 < initial_unit_parameters[1][3] < 1.0

        assert 0.099 < initial_parameters[0][0] < 0.101
        assert 0.099 < initial_parameters[1][0] < 0.101
        assert 0.199 < initial_parameters[0][1] < 0.201
        assert 0.199 < initial_parameters[1][1] < 0.201
        assert 0.299 < initial_parameters[0][2] < 0.301
        assert 0.299 < initial_parameters[1][2] < 0.301
        assert 0.399 < initial_parameters[0][3] < 0.401
        assert 0.399 < initial_parameters[1][3] < 0.401

        assert initial_figures_of_merit == [-1.0e99, -1.0e99]


class TestInitializeBall:
    def test__ball__initial_samples_sample_centre_of_priors(self):

        model = af.PriorModel(MockClassx4)
        model.one = af.UniformPrior(lower_limit=0.0, upper_limit=1.0)
        model.two = af.UniformPrior(lower_limit=0.0, upper_limit=2.0)
        model.three = af.UniformPrior(lower_limit=0.0, upper_limit=3.0)
        model.four = af.UniformPrior(lower_limit=0.0, upper_limit=4.0)

        initializer = af.InitializerBall(lower_limit=0.4999, upper_limit=0.5001)

        initial_unit_parameters, initial_parameters, initial_figures_of_merit = initializer.initial_samples_from_model(
            total_points=2, model=model, fitness_function=MockFitness()
        )

        assert 0.4999 < initial_unit_parameters[0][0] < 0.5001
        assert 0.4999 < initial_unit_parameters[1][0] < 0.5001
        assert 0.4999 < initial_unit_parameters[0][1] < 0.5001
        assert 0.4999 < initial_unit_parameters[1][1] < 0.5001
        assert 0.4999 < initial_unit_parameters[0][2] < 0.5001
        assert 0.4999 < initial_unit_parameters[1][2] < 0.5001
        assert 0.4999 < initial_unit_parameters[0][3] < 0.5001
        assert 0.4999 < initial_unit_parameters[1][3] < 0.5001

        assert 0.499 < initial_parameters[0][0] < 0.501
        assert 0.499 < initial_parameters[1][0] < 0.501
        assert 0.999 < initial_parameters[0][1] < 1.001
        assert 0.999 < initial_parameters[1][1] < 1.001
        assert 1.499 < initial_parameters[0][2] < 1.501
        assert 1.499 < initial_parameters[1][2] < 1.501
        assert 1.999 < initial_parameters[0][3] < 2.001
        assert 1.999 < initial_parameters[1][3] < 2.001

        initializer = af.InitializerBall(lower_limit=0.7999, upper_limit=0.8001)

        initial_unit_parameters, initial_parameters, initial_figures_of_merit = initializer.initial_samples_from_model(
            total_points=2, model=model, fitness_function=MockFitness()
        )

        assert 0.799 < initial_parameters[0][0] < 0.801
        assert 0.799 < initial_parameters[1][0] < 0.801
        assert 1.599 < initial_parameters[0][1] < 1.601
        assert 1.599 < initial_parameters[1][1] < 1.601
        assert 2.399 < initial_parameters[0][2] < 2.401
        assert 2.399 < initial_parameters[1][2] < 2.401
        assert 3.199 < initial_parameters[0][3] < 3.201
        assert 3.199 < initial_parameters[1][3] < 3.201

        assert initial_figures_of_merit == 2 * [1.0]



from collections import ChainMap
from itertools import chain
from functools import reduce
from typing import (
    Dict, Tuple, Optional, NamedTuple, Iterator, List, Union
)
from functools import partial

import numpy as np

from autofit.graphical.factor_graphs import \
    Factor, AbstractNode, FactorGraph, FactorValue, JacobianValue
from autofit.graphical.messages import \
    AbstractMessage, FixedMessage, map_dists
from autofit.graphical.utils import \
    prod, add_arrays, OptResult, Status, aggregate, diag, Axis
from autofit.mapper.variable import Variable

VariableFactorDist = Dict[str, Dict[Factor, AbstractMessage]]
Projection = Dict[str, AbstractMessage]


def project_on_to_factor_approx(
        factor_approx: "FactorApproximation",
        model_dist: Dict[str, AbstractMessage],
        delta: float = 1.,
        status: Optional[Status] = None
) -> Tuple["FactorApproximation", Status]:
    """
    For a passed FactorApproximation this calculates the 
    factor messages such that 
    
    model_dist = factor_dist * cavity_dist
    """
    success, messages = Status() if status is None else status
    assert 0 < delta <= 1

    factor_projection = {}
#     log_norm = 0.
    for v, q_fit in model_dist.items():
        q_cavity = factor_approx.cavity_dist.get(v)
        if isinstance(q_fit, FixedMessage):
            factor_projection[v] = q_fit
        elif q_fit.is_valid:
            if q_cavity:
                q_f0 = factor_approx.factor_dist[v]
                q_f1 = (q_fit / q_cavity)
            else:
                # In the case that q_cavity does not exist the model fit
                # equals the factor approximation
                q_f1 = q_fit

            # weighted update
            if delta != 1:
                q_f1 = (q_f1 ** delta).sum_natural_parameters(q_f0 ** (1 - delta))

            if not q_f1.is_valid:
                # partial updating of values
                q_f1 = q_f1.update_invalid(q_f0)
                messages += (
                    f"factor projection for {v} with {factor_approx.factor} contained "
                    "invalid values",)

            if not q_f1.is_valid:
                success = False
                messages += (
                    f"factor projection for {v} with {factor_approx.factor} is invalid",)

            factor_projection[v] = q_f1
        else:
            success = False
            messages += (
                f"model projection for {v} with {factor_approx.factor} is invalid",)

            factor_projection[v] = factor_approx.factor_dist[v]
            q_model = (q_fit ** delta).sum_natural_parameters(
                factor_approx.model_dist[v] ** (1 - delta))
            if q_model.is_valid:
                model_dist[v] = q_model

    projection = FactorApproximation(
        factor_approx.factor,
        factor_approx.cavity_dist, 
        factor_dist=MeanField(factor_projection),
        model_dist=MeanField(model_dist),
#         log_norm=log_norm
    )
    status = Status(success, messages)

    return projection, status

    
class MeanField(Dict[Variable, AbstractMessage], Factor):
    """For a factor with multiple variables, this class represents the 
    the mean field approximation to that factor, 

    f(x₁, x₂, x₃) = q(x₁, x₂, x₃) = q₁(x₁) q₂(x₂) q₃(x₃)

    Internally these variables approximations are stored in a 
    dictionary with the variables as keys and the message or 
    variable distribution as values


    Methods
    -------
    keys()
        returns the variables of the mean_field

    logpdf({x₁: x1, x₂: x2, x₃: x3})
        returns the q(x₁, x₂, x₃), axis defines the axes over which
        to reduce the return, if the meanfield is duplicated over multiple
        plates. 

    logpdf_gradient({x₁: x1, x₂: x2, x₃: x3})
        returns the q(x₁, x₂, x₃) and the gradients for each input. 
        to save memory the gradients are always the shape of the input
        values (i.e. this does not calculate the Jacobian)
    """
    def __init__(
            self, 
            dists: Dict[Variable, AbstractMessage], 
            log_norm: np.ndarray = 0.):
        dict.__init__(self, dists)
        Factor.__init__(
            self, self._logpdf, **{v.name: v for v in dists})

        if isinstance(dists, MeanField):
            self.log_norm = dists.log_norm
        else:
            self.log_norm = log_norm

    pop = dict.pop 
    values = dict.values 
    items = dict.items
    __getitem__ = dict.__getitem__
        
    def _logpdf(self, **kwargs: np.ndarray) -> np.ndarray:
        var_names = self.name_variable_dict
        return self.logpdf(
            {var_names[k]: value for k, value in kwargs.items()})

    @property
    def mean(self):
        return {v: dist.mean for v, dist in self.items()}

    @property 
    def variance(self):
        return {v: dist.variance for v, dist in self.items()}

    @property 
    def scale(self):
        return {v: dist.scale for v, dist in self.items()}
    
    def logpdf(
            self, 
            values: Dict[Variable, np.ndarray],
            axis: Axis = False, 
    ) -> np.ndarray:
        """Calculates the logpdf of the passed values for messages 

        the result is broadcast to the appropriate shape given the variable
        plates
        """
        return reduce(
            add_arrays, 
            (aggregate(
                self._broadcast(
                    self._variable_plates[v], m.logpdf(values[v])),
                axis = axis)
            for v, m in self.items())
        )

    def __call__(
            self, 
            values: Dict[Variable, np.ndarray],
            axis: Axis = False, 
    ) -> FactorValue:
        return FactorValue(self.logpdf(values, axis=axis), {})

    def logpdf_gradient(
            self, 
            values: Dict[Variable, np.ndarray], 
            axis: Axis = False, 
            **kwargs):
        logl = 0
        gradl = {}
        for v, m in self.items():
            lv, gradl[v] = m.logpdf_gradient(values[v])
            lv = aggregate(
                self._broadcast(self._variable_plates[v], lv),
                axis = axis)
            logl = add_arrays(logl, lv)

        return logl, gradl

    def logpdf_gradient_hessian(
            self, 
            values: Dict[Variable, np.ndarray], 
            axis: Optional[Union[bool, int, Tuple[int, ...]]] = False, 
            **kwargs):
        logl = 0.
        gradl = {}
        hessl = {}
        for v, m in self.items():
            lv, gradl[v], hessl[v] = m.logpdf_gradient_hessian(values[v])
            lv = aggregate(
                self._broadcast(self._variable_plates[v], lv),
                axis = axis)
            logl = add_arrays(logl, lv)

        return logl, gradl, hessl

    def __repr__(self):
        reprdict = "{\n" + "\n".join(
            "  {}: {}".format(k, v) for k, v in self.items()) + "\n  }"
        classname = (type(self).__name__)
        return f"{classname}({reprdict}, log_norm={self.log_norm})"
    
    @property
    def is_valid(self):
        return all(d.is_valid for d in self.values())
    
    def prod(self, *approxs: 'MeanField') -> 'MeanField':
        dists = (
            (k, prod((m.get(k, 1.) for m in approxs), m))
            for k, m in self.items())
        return MeanField({
            k: m for k, m in dists if isinstance(m, AbstractMessage)})

    __mul__ = prod
    
    def __truediv__(self, other: 'MeanField') -> 'MeanField':
        return type(self)({
            k: m / other.get(k, 1.) for k, m in self.items()},
            self.log_norm - other.log_norm)

    def __pow__(self, other: float) -> 'MeanField':
        return type(self)({
            k: m**other for k, m in self.items()},
            self.log_norm * other)

    def log_normalisation(self, other: 'MeanField') -> float:
        return sum(
            np.sum(dist.log_normalisation(other[v]))
            for v, dist in self.items()
        )

    def update_invalid(self, other: "MeanField") -> "MeanField":
        mean_field = {}
        for k, m in self.items():
            m2 = other.get(k)
            mean_field[k] = m.update_invalid(m2) if m2 else m

        return type(self)(mean_field, self.log_norm)

    def project_mode(self, res: OptResult):
        projection = type(self)({
            v: dist.from_mode(res.mode[v], res.hess_inv.get(v))
            for v, dist in self.items()})
        
        projection.log_norm = (
            res.log_norm - projection(res.mode, axis=None).log_value)
        return projection

    def _project_mode(
            self, 
            mode: Dict[Variable, np.ndarray],
            covar: Dict[Variable, np.ndarray], 
            fun: Optional[float] = None):
        """
        Projects the mode and covariance 
        """
        projection = MeanField({
            v: dist.from_mode(mode[v], covar.get(v))
            for v, dist in self.items()})
        if fun is not None:
            projection.log_norm = fun - projection(mode).log_value
            
        return projection

    def sample(self, n_samples=None):
        return {v: dist.sample(n_samples) for v, dist in self.items()}

    def kl(self, mean_field: 'MeanField') -> np.ndarray:
        return sum(
            np.sum(dist.kl(mean_field[k]))
            for k, dist in self.items()
        ) 

    __hash__ = Factor.__hash__ 
    
    @classmethod
    def from_dist(
        cls, 
        dist: Union[Dict[Variable, AbstractMessage], "MeanField"]
    ) -> "MeanField":
        return dist if isinstance(dist, cls) else MeanField(dist)


class FactorApproximation(AbstractNode):
    """
    This class represents the 'tilted distribution' in EP,

    When approximating a model distribution of the form,

    m(x) = ∏ₐ fₐ(xₐ)

    we can define an approximating distribution as the product of 
    factor distributions,

    q(x) = ∏ₐ qₐ(xₐ)

    the 'cavity distribution' q⁻ᵃ for a factor can be viewed as a 
    prior distribution for the factor,

    q⁻ᵃ(xₐ) = ∏_{ᵦ ≠ a} qᵦ(xᵦ)

    so the model can be approximated by the 'tilted distribution'

    q⁺ᵃ(xₐ) = fₐ(xₐ) q⁻ᵃ(xₐ)

    Parameters
    ----------
    is_valid
    factor: Factor
        fₐ(xₐ)
    cavity_dist: MeanField
        q⁻ᵃ(xₐ)
    factor_dist: MeanField
        qₐ(xₐ)
    model_dist: MeanField
        q(xₐ)

    Methods
    -------
    __call__(values={xₐ: x₀}, axis=axis)
        returns q⁺ᵃ(x₀)

    func_jacobian(values={xₐ: x₀}, variables=(xₐ,), axis=axis)
        returns q⁺ᵃ(x₀), {xₐ: dq⁺ᵃ(x₀)/dxₐ}

    project_mean_field(mean_field, delta=1., status=Status())
        for qᶠ = mean_field, finds qₐ such that qᶠₐ * q⁻ᵃ = qᶠ
        delta controls how much to change from the original factor qₐ
        so qʳₐ = (qᶠₐ)ᵟ * (qᶠₐ)¹⁻ᵟ

        returns qʳₐ, status
    """
    def __init__(
        self, 
        factor: Factor, 
        cavity_dist: MeanField, 
        factor_dist: MeanField, 
        model_dist: MeanField
    ):
        # Have to seperate FactorApproximation into two classes
        # in order to be able to redefine __new__
        self.factor = factor
        self.cavity_dist = MeanField.from_dist(cavity_dist)
        self.factor_dist = MeanField.from_dist(factor_dist)
        self.model_dist = MeanField.from_dist(model_dist)

        super().__init__(**factor._kwargs)

    @property
    def variables(self):
        return self.factor.variables

    @property
    def deterministic_variables(self):
        return self.factor.deterministic_variables

    @property
    def all_variables(self):
        return self.factor.all_variables

    @property
    def name(self):
        return f"FactorApproximation({self.factor.name})"

    @property
    def deterministic_dist(self):
        """
        the `MeanField` approximation of the deterministic variables
        """
        return MeanField({
            v: self.cavity_dist[v] for v in self.deterministic_variables})

    @property
    def is_valid(self) -> bool:
        """
        returns whether all the distributions in the factor approximation
        are valid
        """
        dists = chain(
            self.cavity_dist.values(),
            self.factor_dist.values(),
            self.model_dist.values())
        return all(d.is_valid for d in dists if isinstance(d, AbstractMessage))

    def __call__(
            self, 
            values: Dict[Variable, np.ndarray],
            axis: Axis = False, 
    ) -> FactorValue:
        fval = self.factor(values, axis=axis)
        log_meanfield = self.cavity_dist(
            {**values, **fval.deterministic_values}, axis=axis)
        return add_arrays(fval, log_meanfield)

    def func_jacobian(
            self, 
            variable_dict: Dict[Variable, np.ndarray],
            variables: Optional[List[Variable]] = None,
            axis: Axis = None,
            _calc_deterministic: bool = True,
            **kwargs, 
    ) -> Tuple[FactorValue, JacobianValue]:

        if axis is not None:
            raise NotImplementedError(
                "FactorApproximation.func_jacobian has not implemeted "
                f"axis={axis}, try axis=None")

        if variables is None:
            fixed_variables = set(
                v for v, m in self.model_dist.items() 
                if isinstance(m, FixedMessage))
            variables = self.factor.variables - fixed_variables

        fval, fjac = self.factor.func_jacobian(
            variable_dict, variables, axis=axis, 
            _calc_deterministic=_calc_deterministic)

        values = {**variable_dict, **fval.deterministic_values}
        var_sizes = {v: np.size(x) for v, x in values.items()}
        var_shapes = {v: np.shape(x) for v, x in values.items()}
        log_cavity, grad_cavity = self.cavity_dist.logpdf_gradient(
            values, axis=axis)

        logl = fval + log_cavity

        for v in fjac:
            fjac[v] += grad_cavity[v]

        # Update gradients using jacobians of deterministic variables
        # TODO: Should add logic to account for pullbacks for 
        #       AD frameworks e.g. Zygote.jl
        for var, grad in fjac.items():
            for det, jac in grad.deterministic_values.items():
                det_grad = grad_cavity[det].ravel()
                g = jac.reshape(var_sizes[det], var_sizes[var])
                fjac[var] += det_grad.dot(g).reshape(var_shapes[var])
        
        return logl, fjac

    def project_mean_field(
            self, 
            model_dist: MeanField, 
            delta: float = 1.,
            status: Optional[Status] = None,
    ) -> "FactorApprox":
        success, messages = Status() if status is None else status

        factor_dist = (model_dist / self.cavity_dist)
        if delta < 1:
            log_norm = factor_dist.log_norm
            factor_dist = (
                factor_dist**delta * self.factor_dist**(1-delta))
            factor_dist.log_norm = (
                delta * log_norm + (1 - delta) *  self.factor_dist.log_norm)

        if not factor_dist.is_valid:
            success = False
            messages += (
                f"model projection for {self} is invalid",)
            factor_dist = factor_dist.update_invalid(self.factor_dist)

        new_approx = FactorApproximation(
            self.factor,
            self.cavity_dist, 
            factor_dist=factor_dist,
            model_dist=model_dist,
        )
        return new_approx, Status(success, messages)

    project = project_mean_field

    def __repr__(self):
        # TODO make this nicer
        return f"{type(self).__name__}({self.factor}, ...)"




from itertools import repeat, chain
from typing import \
    Tuple, Dict, List, Callable, Optional, Union
from functools import reduce 

import numpy as np

from autofit.mapper.variable import Variable
from autofit.graphical.factor_graphs.abstract import \
    FactorValue, JacobianValue
from autofit.graphical.factor_graphs.factor import \
    AbstractFactor, Factor, DeterministicFactor
from autofit.graphical.utils import \
    aggregate, Axis, cached_property


class FactorJacobian(Factor):
    """
    A node in a graph representing a factor with analytic evaluation 
    of its Jacobian

    Parameters
    ----------
    factor_jacobian
        the function being wrapped, which can also return its jacobian by 
        using the keyword argument `_variables` which indicates which, 
        if any, variables the function must return the Jacobians for, 
        so for example, for a function with variables 'x' and 'y', the 
        call signatures would look like so, 

        z, (dz_dx, dz_dy) = factor_jacoban(x=x, y=y, _variables=('x', 'y'))
        z, (dz_dy, dz_dx) = factor_jacoban(x=x, y=y, _variables=('y', 'x'))
        z, (dz_dy,) = factor_jacoban(x=x, y=y, _variables=('y',))
        z, () = factor_jacoban(x=x, y=y, _variables=())
        z = factor_jacoban(x=x, y=y, _variables=None)

    name: optional, str
        the name of the factor, if not passed then uses the name
        of the function passed

    vectorised: optional, bool
        if true the factor will call it the function directly over multiple 
        inputs. If false the factor will call the function iteratively over
        each argument. 

    is_scalar: optional, bool
        if true the factor returns a scalar value. Note if multiple arguments
        are passed then a vector will still be returned

    kwargs: Variables
        Variables for each keyword argument for the function
    """
    def __init__(
            self,
            factor_jacobian: Callable,
            *, 
            name="",
            vectorised=False,
            is_scalar=False, 
            **kwargs: Variable
    ):
        self.vectorised = vectorised
        self.is_scalar = is_scalar
        self._factor = factor_jacobian
        AbstractFactor.__init__(
            self, 
            **kwargs,
            name=name or factor_jacobian.__name__
        )

    def __hash__(self) -> int:
        # TODO: might this break factor repetition somewhere?
        return hash(self._factor)

    def _call_factor(
            self,
            values: Dict[str, np.ndarray],
            variables: Optional[Tuple[str, ...]] = None, 
    ) -> Union[np.ndarray, Tuple[np.ndarray, Tuple[np.ndarray, ...]]]:
        """
        Call the underlying function

        Parameters
        ----------
        args
            Positional arguments for the function
        kwargs
            Keyword arguments for the function

        Returns
        -------
        Value returned by the factor
        """
        if self.vectorised:
            return self._factor(**values, _variables=variables)
            
        """Some factors may not be vectorised to broadcast over
        multiple inputs

        this method checks whether multiple input values have been
        passed, and if so automatically loops over the inputs.
        If any of the inputs have initial dimension one, it repeats
        that value to match the length of the other inputs

        If the other inputs do not match then it raises ValueError
        """
        kwargs_dims = {k: np.ndim(a) for k, a in values.items()}
        # Check dimensions of inputs directly match plates
        direct_call = (
            all(dim == kwargs_dims[k] for k, dim in self._kwargs_dims.items()))
        if direct_call:
            return self._factor(**values, _variables=variables)
        else:
            return self._multicall_factor(values, variables)

    def _multicall_factor(
            self,
            values: Dict[str, np.ndarray],
            variables: Optional[Tuple[str, ...]] = None, 
    ) -> Union[np.ndarray, Tuple[np.ndarray, Tuple[np.ndarray, ...]]]:
        """call the factor multiple times and aggregates 
        the results together
        """
        # Check dimensions of inputs match plates + 1
        vectorised = all(
            dim + 1 == np.ndim(values[k]) 
            for k, dim in self._kwargs_dims.items())

        if not vectorised:
            kwargs_dims = {k: np.ndim(a) for k, a in values.items()}
            raise ValueError(
                "input dimensions do not match required dims"
                f"input: **kwargs={kwargs_dims}"
                f"required: "
                f"**kwargs={self._kwargs_dims}")

        kw_lens = {k: len(a) for k, a in values.items()}

        # checking 1st dimensions match
        sizes = set(kw_lens.values())
        dim0 = max(sizes)
        if sizes.difference({1, dim0}):
            raise ValueError(
                f"size mismatch first dimensions passed: {sizes}")

        iter_kws = {
            k: iter(a) if kw_lens[k] == dim0 else iter(repeat(a[0]))
            for k, a in values.items()}

        # TODO this loop can also be parallelised for increased performance
        fjacs = [
            self._factor(**{
                    k: next(a) for k, a in iter_kws.items()}, 
                    _variables=variables)
             for _ in range(dim0)]
        
        if variables is None:
            res = np.array([fjac for fjac in fjacs])
            return res 
        else:
            res = np.array([fjac[0] for fjac in fjacs])
            njac = len(fjacs[0][1])
            jacs = tuple(
                np.array([fjac[1][i] for fjac in fjacs])
                for i in range(njac))

            return res, jacs

    def __call__(
            self,
            variable_dict: Dict[Variable, np.ndarray],
            axis: Axis = False, 
    ) -> FactorValue:
        values = self.resolve_variable_dict(variable_dict)
        val = self._call_factor(values, variables=None)
        val = aggregate(val, axis)
        return FactorValue(val, {})

    def func_jacobian(
            self,
            variable_dict: Dict[Variable, np.ndarray],
            variables: Optional[Tuple[Variable, ...]] = None,
            axis: Axis = False, 
            **kwargs
    ) -> Tuple[FactorValue, JacobianValue]:
        """
        Call the underlying factor

        Parameters
        ----------
        variable_dict : 
            the values to call the function with
        variables : tuple of Variables
            the variables to calculate gradients and Jacobians for
            if variables = None then differentiates wrt all variables
        axis : None or False or int or tuple of ints, optional
            the axes to reduce the result over.
            if axis = None the sum over all dimensions 
            if axis = False then does not reduce result

        Returns
        -------
        FactorValue, JacobianValue
            encapsulating the result of the function call
        """
        if variables is None:
            variables = self.variables

        variable_names = tuple(
            self._variable_name_kw[v.name]
            for v in variables)
        kwargs = self.resolve_variable_dict(variable_dict)
        val, jacs = self._call_factor(
            kwargs, variables=variable_names)
        grad_axis = tuple(range(np.ndim(val))) if axis is None else axis
        
        fval = FactorValue(
            aggregate(self._reshape_factor(val, kwargs), axis))
        fjac = {
            v: FactorValue(aggregate(jac, grad_axis))
            for v, jac in zip(variables, jacs)
        }
        return fval, fjac

    def __eq__(self, other: Union["Factor", Variable]):
        """
        If set equal to a variable that variable is taken to be deterministic and
        so a DeterministicFactorNode is generated.
        """
        if isinstance(other, Factor):
            if isinstance(other, type(self)):
                return (
                        (self._factor == other._factor)
                        and (frozenset(self._kwargs.items())
                             == frozenset(other._kwargs.items()))
                        and (frozenset(self.variables)
                             == frozenset(other.variables))
                        and (frozenset(self.deterministic_variables)
                             == frozenset(self.deterministic_variables)))
            else:
                return False

        return DeterministicFactorJacobian(
            self._factor,
            other,
            **self._kwargs
        )

    def __repr__(self) -> str:
        args = ", ".join(chain(
            map("{0[0]}={0[1]}".format, self._kwargs.items())))
        return f"FactorJacobian({self.name}, {args})"


class DeterministicFactorJacobian(FactorJacobian):
    """
    A deterministic factor is used to convert a function f(g(x)) to f(y)g(x) (integrating over y wit
    a delta function) so that it can be represented in a factor graph.

    Parameters
    ----------
    factor
        The original factor to which the deterministic factor is associated
    variable
        The deterministic variable that is returned by the factor, so 
        to represent the case f(g(x)), we would define,

        ```
        >>> x = Variable('x')
        >>> y = Variable('y')
        >>> g_ = Factor(g, x) == y
        >>> f_ = Factor(f, y)
        ```
        Alternatively g could be directly defined,
        ```
        >>> g_ = DeterministicFactor(g, y, x=x)
        ```
        
    kwargs
        Variables for the original factor
    """
    def __init__(
            self,
            factor_jacobian: Callable,
            variable: Variable,
            vectorised=False,
            is_scalar=False,
            **kwargs: Variable
    ):
        
        super().__init__(
            factor_jacobian,
            vectorised=vectorised,
            is_scalar=is_scalar, 
            **kwargs
        )
        self._deterministic_variables = variable, 

    @property
    def deterministic_variables(self):
        return set(self._deterministic_variables)

    def func_jacobian(
            self,
            variable_dict: Dict[Variable, np.ndarray],
            variables: Optional[Tuple[Variable, ...]] = None,
            axis: Axis = False, 
            **kwargs, 
    ) -> Tuple[FactorValue, JacobianValue]:
        """
        Call this factor with a set of arguments

        Parameters
        ----------
        args
            Positional arguments for the underlying factor
        kwargs
            Keyword arguments for the underlying factor

        Returns
        -------
        An object encapsulating the value for the factor
        """
        if variables is None:
            variables = self.variables

        variable_names = tuple(
            self._variable_name_kw[v.name] for v in variables)
        kwargs = self.resolve_variable_dict(variable_dict)
        vals, *jacs = self._call_factor(
            kwargs, variables=variable_names)
        shift, shape = self._function_shape(**kwargs)
        plate_dim = dict(zip(self.plates, shape[shift:]))


        det_shapes = {
            v: shape[:shift] + tuple(
                plate_dim[p] for p in v.plates)
            for v in self.deterministic_variables
        }
        var_shapes = {
            self._kwargs[v]: np.shape(x) for v, x in kwargs.items()}

        if not (isinstance(vals, tuple) or self.n_deterministic > 1):
            vals = vals,

        log_val = (
            0. if (shape == () or axis is None) else 
            aggregate(np.zeros(tuple(1 for _ in shape)), axis))
        det_vals = {
            k: np.reshape(val, det_shapes[k])
            if det_shapes[k] else val
            for k, val in zip(self._deterministic_variables, vals)
        }
        fval = FactorValue(log_val, det_vals)

        vjacs = {}
        for k, _jacs in zip(self._deterministic_variables, jacs):
            for v, jac in zip(variables, _jacs):
                vjacs.setdefault(v, {})[k] = np.reshape(
                    jac, det_shapes[k] + var_shapes[v][shift:])
        fjac = {
            v: FactorValue(
                np.zeros(np.shape(log_val) + var_shapes[v]),
                vjacs[v])
            for v in variables
        }
        return fval, fjac

    def __call__(
            self,
            variable_dict: Dict[Variable, np.ndarray],
            axis: Axis = False, 
    ) -> FactorValue:
        return self.func_jacobian(variable_dict, (), axis=axis)[0]

    def __repr__(self) -> str:
        factor_str = super().__repr__()
        var_str = ", ".join(sorted(variable.name for variable in self._deterministic_variables))
        return f"({factor_str} == ({var_str}))"


from os import path
import pytest

import autofit as af
from autofit.mock.mock import MockSearchOutput


def test_completed_aggregator(aggregator_directory):

    print(aggregator_directory)

    aggregator = af.Aggregator(aggregator_directory, completed_only=True)
    assert len(aggregator) == 1
    assert "completed" in aggregator[0].directory


class TestLoading:
    def test_unzip(self, path_aggregator):
        assert len(path_aggregator) == 2

    def test_pickles(self, path_aggregator):
        assert list(path_aggregator.values("dataset"))[0]["name"] == "dataset"
        assert list(path_aggregator.values("model"))[0]["name"] == "model"
        assert list(path_aggregator.values("non_linear"))[0]["name"] == "optimizer"
        assert list(path_aggregator.values("nonsense"))[0] is None


@pytest.fixture(name="aggregator_2")
def make_aggregator_2():
    aggregator = af.Aggregator("")
    aggregator.search_outputs = [
        MockSearchOutput(path.join("directory", "number", "one"), "apipeline1", "search1", "dataset1"),
        MockSearchOutput(path.join("directory", "number", "two"), "pipeline1", "bsearch2", "dataset1"),
        MockSearchOutput(path.join("directory", "letter", "a"), "pipeline2", "search2", "dataset2"),
    ]
    return aggregator


class TestIntersection:

    def test_on_pipeline(self, aggregator, aggregator_2):
        aggregator, aggregator_2 = aggregator.homogenize(aggregator_2, on="pipeline")
        assert list(aggregator.values("search")) == ["search1", "search2", "search2"]
        assert list(aggregator_2.values("search")) == ["bsearch2", "search2"]

    def test_on_search(self, aggregator, aggregator_2):
        aggregator, aggregator_2 = aggregator.homogenize(aggregator_2, on="search")
        assert list(aggregator.values("pipeline")) == [
            "pipeline1",
            "pipeline1",
            "pipeline2",
        ]
        assert list(aggregator_2.values("pipeline")) == ["apipeline1", "pipeline2"]


class TestOperations:
    def test_attribute(self, aggregator):
        assert list(aggregator.values("pipeline")) == [
            "pipeline1",
            "pipeline1",
            "pipeline2",
        ]
        assert list(aggregator.values("search")) == ["search1", "search2", "search2"]
        assert list(aggregator.values("dataset")) == [
            "dataset1",
            "dataset1",
            "dataset2",
        ]

    def test_indexing(self, aggregator):
        assert list(aggregator[1:].values("pipeline")) == ["pipeline1", "pipeline2"]
        assert list(aggregator[-1:].values("pipeline")) == ["pipeline2"]
        assert aggregator[0].pipeline == "pipeline1"

    def test_group_by(self, aggregator):
        result = aggregator.group_by("pipeline")

        assert len(result) == 2
        assert len(result[0]) == 2
        assert len(result[1]) == 1

        result = result.filter(aggregator.search == "search2")

        assert len(result) == 2
        assert len(result[0]) == 1
        assert len(result[1]) == 1

        assert list(map(list, result.values("search"))) == [["search2"], ["search2"]]

    def test_map(self, aggregator):
        def some_function(output):
            return f"{output.search} {output.dataset}"

        results = aggregator.map(some_function)
        assert list(results) == [
            "search1 dataset1",
            "search2 dataset1",
            "search2 dataset2",
        ]



from codecs import open
from os.path import abspath, dirname, join
from os import environ

from setuptools import find_packages, setup

this_dir = abspath(dirname(__file__))
with open(join(this_dir, "README.rst"), encoding="utf-8") as file:
    long_description = file.read()

with open(join(this_dir, "requirements.txt")) as f:
    requirements = f.read().split("\n")

requirements.extend([
    'autoconf==0.9.0'
])

setup(
    name="autofit",
    version=environ.get("VERSION", "0.75.0"),
    description="Classy Probabilistic Programming",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rhayes777/AutoFit",
    author="James Nightingale and Richard Hayes",
    author_email="richard@rghsoftware.co.uk",
    include_package_data=True,
    license="MIT License",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Physics",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords="cli",
    packages=find_packages(exclude=["docs"]),
    install_requires=requirements,
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
)



from abc import ABC, abstractmethod
from collections import defaultdict
from itertools import count
from typing import (
    Dict, Tuple, Optional, List,
    Callable
)

import numpy as np

from autofit.graphical.factor_graphs import (
    Factor, AbstractNode, FactorGraph
)
from autofit.graphical.mean_field import MeanField, FactorApproximation
from autofit.graphical.messages.abstract import AbstractMessage
from autofit.graphical.utils import Status
from autofit.mapper.variable import Variable


class EPMeanField(FactorGraph):
    '''
    this class encode the EP mean-field approximation to a factor graph

    
    Attributes
    ----------
    factor_graph: FactorGraph
        the base factor graph being approximated

    factor_mean_field: Dict[Factor, MeanField]
        the mean-field approximation for each factor in the factor graph

    mean_field: MeanField
        the mean-field approximation of the full factor graph
        i.e. the product of the factor mean-field approximations

    variables: Set[Variable]
        the variables of the approximation

    deterministic_variables: Set[Variable]
        the deterministic variables

    log_evidence: float
        the approximate log evidence of the approximation

    is_valid: bool
        returns whether the factor mean-field approximations are all valid

    Methods
    -------
    from_approx_dists(factor_graph, approx_dists)
        create a EPMeanField object from the passed factor_graph
        using approx_dists to initialise the factor mean-field approximations

    factor_approximation(factor)
        create the FactorApproximation for the factor

    project_factor_approx(factor_approximation)
        given the passed FactorApproximation, return a new `EPMeanField`
        object encoding the updated mean-field approximation
    '''

    def __init__(
            self,
            factor_graph: FactorGraph,
            factor_mean_field: Dict[Factor, MeanField]
    ):
        self._factor_graph = factor_graph
        self._factor_mean_field = factor_mean_field
        variable_factor = {}
        for factor, vs in factor_graph.factor_all_variables.items():
            for v in vs:
                variable_factor.setdefault(v, set()).add(factor)
        self._variable_factor = variable_factor

        super().__init__(self.factor_graph.factors)

    @property
    def name(self):
        return f"EP_{self.factor_graph.name}"

    @property
    def variables(self):
        return self.factor_graph.variables

    @property
    def deterministic_variables(self):
        return self.factor_graph.deterministic_variables

    # @property
    # def variable_names(self) -> Dict[str, Variable]: 
    #     return self.factor_graph.variable_names

    @property
    def factor_mean_field(self) -> Dict[Factor, MeanField]:
        return self._factor_mean_field.copy()

    @property
    def factor_graph(self) -> FactorGraph:
        return self._factor_graph

    @classmethod
    def from_approx_dists(
            cls,
            factor_graph: FactorGraph,
            approx_dists: Dict[Variable, AbstractMessage],
    ) -> "EPMeanField":
        factor_mean_field = {
            factor: MeanField({
                v: approx_dists[v] for v in factor.all_variables})
            for factor in factor_graph.factors}

        return cls(
            factor_graph,
            factor_mean_field)

    from_kws = from_approx_dists

    def factor_approximation(self, factor: Factor) -> FactorApproximation:
        factor_mean_field = self._factor_mean_field.copy()
        factor_dist = factor_mean_field.pop(factor)
        cavity_dist = MeanField.prod(
            {v: 1. for v in factor_dist.all_variables},
            *(dist for fac, dist in factor_mean_field.items()))
        # cavity_dist.log_norm = 0.
        model_dist = factor_dist.prod(cavity_dist)

        return FactorApproximation(
            factor, cavity_dist, factor_dist, model_dist)

    def project_factor_approx(
            self, projection: FactorApproximation, status: Optional[Status] = None,
    ) -> "EPMeanField":
        """
        """
        factor_mean_field = self.factor_mean_field
        factor_mean_field[projection.factor] = projection.factor_dist

        new_approx = type(self)(
            factor_graph=self._factor_graph,
            factor_mean_field=factor_mean_field)
        return new_approx, status

    project = project_factor_approx

    @property
    def mean_field(self) -> MeanField:
        return MeanField.prod(
            {v: 1. for v in self.all_variables},
            *self._factor_mean_field.values())

    model_dist = mean_field

    @property
    def variable_factor_message(self
                                ) -> Dict[Variable, Dict[Factor, AbstractMessage]]:
        variable_factor_message = {
            v: {} for v in self.all_variables}
        for factor, meanfield in self.factor_mean_field.items():
            for v, message in meanfield.items():
                variable_factor_message[v][factor] = message

        return variable_factor_message

    @property
    def variable_messages(self) -> Dict[Variable, List[AbstractMessage]]:
        variable_messages = {
            v: [] for v in self.all_variables}
        for meanfield in self.factor_mean_field.values():
            for v, message in meanfield.items():
                variable_messages[v].append(message)

        return variable_messages

    @property
    def variable_evidence(self) -> Dict[Variable, np.ndarray]:
        return {
            v: AbstractMessage.log_normalisation(*ms)
            for v, ms in self.variable_messages.items()
        }

    @property
    def factor_evidence(self) -> Dict[Factor, np.ndarray]:
        return {
            factor: meanfield.log_norm
            for factor, meanfield in self.factor_mean_field.items()
        }

    @property
    def log_evidence(self):
        """
        Calculates evidence for the EP approximation

        Evidence for a variable, xᵢ,

        Zᵢ = ∫ ∏ₐ m_{a → i} (xᵢ) dxᵢ

        Evidence for a factor, f_a,

                ∫ ∏_{j ∈ a} m_{i → a} (xᵢ) fₐ(xₐ) dxₐ
        Zₐ = -----------------------------------------
                             ∏_{j ∈ a} Zⱼ

        Evidence for model

        Z = ∏ᵢ Zᵢ ∏ₐ Zₐ
        """
        variable_evidence = {
            v: np.sum(logz) for v, logz in self.variable_evidence.items()}
        factor_evidence = sum(
            np.sum(meanfield.log_norm)
            - sum(variable_evidence[v] for v in factor.all_variables)
            for factor, meanfield in self.factor_mean_field.items()
        )
        return factor_evidence + sum(variable_evidence.values())

    def __repr__(self) -> str:
        clsname = type(self).__name__
        return (
            f"{clsname}({self.factor_graph}, "
            f"log_evidence={self.log_evidence})")

    @property
    def is_valid(self) -> bool:
        return all(mean_field.is_valid
                   for mean_field in self.factor_mean_field.values())


class AbstractFactorOptimiser(ABC):
    @abstractmethod
    def optimise(
            self,
            factor: Factor,
            model_approx: EPMeanField,
            status: Optional[Status] = None
    ) -> Tuple[EPMeanField, Status]:
        pass


EPCallBack = Callable[[Factor, EPMeanField, Status], bool]


class EPHistory:
    def __init__(
            self,
            callbacks: Tuple[EPCallBack, ...] = (),
            kl_tol=1e-1,
            evidence_tol=None):
        self._callbacks = callbacks
        self.history = {}
        self.statuses = {}
        self.factor_count = defaultdict(count)

        self.kl_tol = kl_tol
        self.evidence_tol = evidence_tol

    def __call__(
            self,
            factor: Factor,
            approx: EPMeanField,
            status: Status = Status()
    ) -> bool:
        i = next(self.factor_count[factor])
        self.history[i, factor] = approx
        self.statuses[i, factor] = status

        stop = any([
            callback(factor, approx, status) for callback in self._callbacks
        ])
        if stop:
            return True
        elif i:
            last_approx = self.history[i - 1, factor]
            return self._check_convergence(approx, last_approx)

        return False

    def _kl_convergence(
            self,
            approx: EPMeanField,
            last_approx: EPMeanField,
    ) -> bool:
        return approx.mean_field.kl(last_approx.mean_field) < self.kl_tol

    def _evidence_convergence(
            self,
            approx: EPMeanField,
            last_approx: EPMeanField,
    ) -> bool:
        last_evidence = last_approx.log_evidence
        evidence = approx.log_evidence
        if last_evidence > evidence:
            # todo print warning?
            return False

        return evidence - last_evidence < self.evidence_tol

    def _check_convergence(
            self,
            approx: EPMeanField,
            last_approx: EPMeanField,
    ) -> bool:
        stop = False
        if self.kl_tol:
            stop = stop or self._kl_convergence(approx, last_approx)

        if self.evidence_tol:
            stop = stop or self._evidence_convergence(approx, last_approx)

        return stop


class EPOptimiser:
    """
    """

    def __init__(
            self,
            factor_graph: FactorGraph,
            default_optimiser: AbstractFactorOptimiser = None,
            factor_optimisers: Dict[Factor, AbstractFactorOptimiser] = None,
            callback: Optional[EPCallBack] = None,
            factor_order: Optional[List[Factor]] = None
    ):
        factor_optimisers = factor_optimisers or {}
        self.factor_graph = factor_graph
        self.factors = factor_order or self.factor_graph.factors

        if default_optimiser is None:
            self.factor_optimisers = factor_optimisers
            missing = set(self.factors) - self.factor_optimisers.keys()
            if missing:
                raise (ValueError(
                    f"missing optimisers for {missing}, "
                    "pass a default_optimiser or add missing optimsers"
                ))
        else:
            self.factor_optimisers = {
                factor: factor_optimisers.get(factor, default_optimiser)
                for factor in self.factors}

        self.callback = callback or EPHistory()

    def run(
            self,
            model_approx: EPMeanField,
            max_steps=100,
    ) -> EPMeanField:
        for _ in range(max_steps):
            for factor, optimiser in self.factor_optimisers.items():
                model_approx, status = optimiser.optimise(factor, model_approx)
                if self.callback(factor, model_approx, status):
                    break  # callback controls convergence
            else:  # If no break do next iteration
                continue
            break  # stop iterations

        return model_approx



import pickle

import autofit as af
from autofit.non_linear.paths import DirectoryPaths


class TestCase:
    def test_simple_pickle(self):
        optimiser = af.DynestyStatic()
        optimiser.paths = DirectoryPaths("name")
        pickled_optimiser = pickle.loads(pickle.dumps(optimiser))
        assert optimiser.paths.path_prefix == pickled_optimiser.paths.path_prefix



from typing import Union
from .abstract import \
    Value, FactorValue, JacobianValue, HessianValue, AbstractNode, Variable, \
    Plate
from .factor import \
    AbstractFactor, Factor, DeterministicFactor
from .jacobians import \
    FactorJacobian, DeterministicFactorJacobian
from .graph import FactorGraph
from .transform import \
    DiagonalTransform, CholeskyTransform, VariableTransform, \
    FullCholeskyTransform, identity_transform, TransformedNode

FactorNode = Union[
    Factor, 
    FactorJacobian
]
DeterministicFactorNode = Union[
    DeterministicFactor,
    DeterministicFactorJacobian
    ]



from os import path


class TestFiltering:
    def test_or(self, aggregator):
        predicate_one = aggregator.directory.contains("one")
        predicate_two = aggregator.directory.contains("two")
        result = aggregator.filter(predicate_one | predicate_two)
        assert len(result) == 2
        assert result.directories == [path.join("directory", "number", "one"), path.join("directory", "number", "two")]

    def test_and(self, aggregator):
        predicate_one = aggregator.pipeline == "pipeline1"
        predicate_two = aggregator.search == "search2"
        result = aggregator.filter(predicate_one & predicate_two)
        assert len(result) == 1
        assert result.directories == [path.join("directory", "number", "two")]

    def test_not_contains(self, aggregator):
        predicate = ~(aggregator.pipeline.contains("1"))
        result = aggregator.filter(predicate)
        assert len(result) == 1
        assert result[0].pipeline == "pipeline2"

    def test_not_equal(self, aggregator):
        predicate = aggregator.pipeline != "pipeline1"
        result = aggregator.filter(predicate)
        assert len(result) == 1
        assert result[0].pipeline == "pipeline2"

    def test_rhs(self, aggregator):
        predicate = "pipeline1" != aggregator.pipeline
        result = aggregator.filter(predicate)
        assert result.pipeline == ["pipeline2"]

        predicate = "pipeline1" == aggregator.pipeline
        result = aggregator.filter(predicate)
        assert result.pipeline == ["pipeline1", "pipeline1"]

    def test_filter_index(self, aggregator):
        assert list(
            aggregator.filter(aggregator.pipeline == "pipeline1")[1:].values("pipeline")
        ) == ["pipeline1"]
        assert list(
            aggregator[1:].filter(aggregator.pipeline == "pipeline1").values("pipeline")
        ) == ["pipeline1"]

    def test_filter(self, aggregator):
        result = aggregator.filter(aggregator.pipeline == "pipeline1")
        assert len(result) == 2
        assert result[0].pipeline == "pipeline1"

        result = aggregator.filter(
            aggregator.pipeline == "pipeline1", aggregator.search == "search1"
        )

        assert len(result) == 1
        assert result[0].pipeline == "pipeline1"

        result = aggregator.filter(aggregator.pipeline == "pipeline1").filter(
            aggregator.search == "search1"
        )

        assert len(result) == 1
        assert result[0].pipeline == "pipeline1"

    def test_filter_contains(self, aggregator):
        result = aggregator.filter(aggregator.pipeline.contains("1"))
        assert len(result) == 2
        assert result[0].pipeline == "pipeline1"

        result = aggregator.filter(
            aggregator.pipeline.contains("1"), aggregator.search.contains("1")
        )

        assert len(result) == 1
        assert result[0].pipeline == "pipeline1"

        result = aggregator.filter(aggregator.pipeline.contains("1")).filter(
            aggregator.search == "search1"
        )

        assert len(result) == 1
        assert result[0].pipeline == "pipeline1"

    def test_filter_contains_directory(self, aggregator):
        result = aggregator.filter(aggregator.directory.contains("number"))
        assert len(result) == 2

        result = aggregator.filter(aggregator.directory.contains("letter"))
        assert len(result) == 1



import os
import re
import shutil
import zipfile
from abc import ABC, abstractmethod
from configparser import NoSectionError
from functools import wraps
from os import path
import json

from autoconf import conf
from autofit.mapper import link
from autofit.mapper.model_object import Identifier
from autofit.non_linear.log import logger
from autofit.text import text_util


def make_path(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        full_path = func(*args, **kwargs)
        os.makedirs(full_path, exist_ok=True)
        return full_path

    return wrapper


pattern = re.compile(r'(?<!^)(?=[A-Z])')


class AbstractPaths(ABC):
    def __init__(
            self,
            name=None,
            path_prefix=None
    ):
        """Manages the path structure for `NonLinearSearch` output, for analyses both not using and using the search
        API. Use via non-linear searches requires manual input of paths, whereas the search API manages this using the
        search attributes.

        The output path within which the *Paths* objects path structure is contained is set via PyAutoConf, using the
        command:

        from autoconf import conf
        conf.instance = conf.Config(output_path="path/to/output")

        If we assume all the input strings above are used with the following example names:

        name = "name"
        path_prefix = "folder_0/folder_1"

        The output path of the `NonLinearSearch` results will be:

        /path/to/output/folder_0/folder_1/name

        Parameters
        ----------
        name : str
            The name of the non-linear search, which is used as a folder name after the ``path_prefix``. For searchs
            this name is the ``name``.
        path_prefix : str
            A prefixed path that appears after the output_path but beflore the name variable.
        """

        self.name = name or ""
        self.path_prefix = path_prefix or ""

        self._search = None
        self.model = None

        self._non_linear_name = None
        self._identifier = None

        try:
            self.remove_files = conf.instance["general"]["output"]["remove_files"]

            if conf.instance["general"]["hpc"]["hpc_mode"]:
                self.remove_files = True
        except NoSectionError as e:
            logger.exception(e)

    @property
    def search(self):
        return self._search

    @search.setter
    def search(self, search):
        self._search = search
        self._non_linear_name = pattern.sub(
            '_', type(
                self.search
            ).__name__
        ).lower()

    @property
    def non_linear_name(self):
        return self._non_linear_name

    @property
    def identifier(self):
        if None in (self.model, self.search):
            logger.warn(
                "Both model and search should be set"
            )
        if self._identifier is None:
            self._identifier = str(
                Identifier([
                    self.search,
                    self.model
                ])
            )
        return self._identifier

    @property
    def path(self):
        return link.make_linked_folder(self._sym_path)

    @property
    @make_path
    def samples_path(self) -> str:
        """
        The path to the samples folder.
        """
        return path.join(self.output_path, "samples")

    @property
    def image_path(self) -> str:
        """
        The path to the image folder.
        """
        return path.join(self.output_path, "image")

    @property
    @make_path
    def output_path(self) -> str:
        """
        The path to the output information for a search.
        """
        strings = (
            list(filter(
                len,
                [
                    str(conf.instance.output_path),
                    self.path_prefix,
                    self.name,
                    self.identifier,
                ],
            )
            )
        )

        return path.join("", *strings)

    def zip_remove(self):
        """
        Copy files from the sym linked search folder then remove the sym linked folder.
        """

        self._zip()

        if self.remove_files:
            try:
                shutil.rmtree(self.path)
            except (FileNotFoundError, PermissionError):
                pass

    def _zip(self):

        try:
            with zipfile.ZipFile(self._zip_path, "w", zipfile.ZIP_DEFLATED) as f:
                for root, dirs, files in os.walk(self.output_path):

                    for file in files:
                        f.write(
                            path.join(root, file),
                            path.join(
                                root[len(self.output_path):], file
                            ),
                        )

            if self.remove_files:
                shutil.rmtree(self.output_path)

        except FileNotFoundError:
            pass

    def restore(self):
        """
        Copy files from the ``.zip`` file to the samples folder.
        """

        if path.exists(self._zip_path):
            with zipfile.ZipFile(self._zip_path, "r") as f:
                f.extractall(self.output_path)

            os.remove(self._zip_path)

    @property
    @make_path
    def _sym_path(self) -> str:
        return path.join(
            conf.instance.output_path,
            self.path_prefix,
            self.name,
            self.identifier,
        )

    def __eq__(self, other):
        return isinstance(other, AbstractPaths) and all(
            [
                self.path_prefix == other.path_prefix,
                self.name == other.name,
                self.non_linear_name == other.non_linear_name,
            ]
        )

    @property
    def _zip_path(self) -> str:
        return f"{self.output_path}.zip"

    @abstractmethod
    def save_object(
            self,
            name: str,
            obj: object
    ):
        pass

    @abstractmethod
    def load_object(
            self,
            name: str
    ):
        pass

    @abstractmethod
    def remove_object(
            self,
            name: str
    ):
        pass

    @abstractmethod
    def is_object(
            self,
            name: str
    ) -> bool:
        pass

    @property
    @abstractmethod
    def is_complete(self) -> bool:
        pass

    @abstractmethod
    def completed(self):
        pass

    @abstractmethod
    def save_all(self, search_config_dict, info, pickle_files):
        pass

    @abstractmethod
    def load_samples(self):
        """
        Load samples from the database
        """

    @abstractmethod
    def save_samples(self, samples):
        """
        Save samples to the database
        """

    @abstractmethod
    def load_samples_info(self):
        pass

    def _save_search(self, config_dict):

        with open(path.join(self.output_path, "search.json"), "w+") as f:
            json.dump(config_dict, f, indent=4)

    def save_summary(self, samples, log_likelihood_function_time):
        text_util.results_to_file(
            samples=samples,
            filename=path.join(
                self.output_path,
                "model.results"
            )
        )

        text_util.search_summary_to_file(
            samples=samples,
            log_likelihood_function_time=log_likelihood_function_time,
            filename=path.join(
                self.output_path,
                "search.summary"
            )
        )

    @property
    def _samples_file(self) -> str:
        return path.join(self.samples_path, "samples.csv")

    @property
    def _info_file(self) -> str:
        return path.join(self.samples_path, "info.json")



from os import path

import pytest

import autofit as af
from autofit.mock.mock import MockClassx4
from autofit.non_linear.samples import NestSamples, Sample

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


@pytest.fixture(name="samples")
def make_samples():
    model = af.ModelMapper(mock_class_1=MockClassx4)

    parameters = [
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [21.0, 22.0, 23.0, 24.0],
        [0.0, 1.0, 2.0, 3.0],
    ]

    return NestSamples(
        model=model,
        samples=Sample.from_lists(
            model=model,
            parameters=parameters,
            log_likelihoods=[1.0, 2.0, 3.0, 10.0, 5.0],
            log_priors=[0.0, 0.0, 0.0, 0.0, 0.0],
            weights=[1.0, 1.0, 1.0, 1.0, 1.0],
        ),
        total_samples=500,
        log_evidence=2,
        unconverged_sample_size=300,
        time=4,
        number_live_points=5,
    )


class TestJsonCSV:
    def test__from_csv_table_and_json_info(self, samples):
        nest = af.DynestyStatic()
        nest.paths = af.DirectoryPaths(path_prefix=path.join("non_linear", "dynesty"))

        samples.write_table(filename=path.join(nest.paths.samples_path, "samples.csv"))
        samples.info_to_json(filename=path.join(nest.paths.samples_path, "info.json"))

        model = af.ModelMapper(mock_class_1=MockClassx4)

        samples = nest.samples_via_csv_json_from_model(model=model)

        assert samples.parameters == [
            [0.0, 1.0, 2.0, 3.0],
            [0.0, 1.0, 2.0, 3.0],
            [0.0, 1.0, 2.0, 3.0],
            [21.0, 22.0, 23.0, 24.0],
            [0.0, 1.0, 2.0, 3.0],
        ]
        assert samples.log_likelihoods == [1.0, 2.0, 3.0, 10.0, 5.0]
        assert samples.log_priors == [0.0, 0.0, 0.0, 0.0, 0.0]
        assert samples.log_posteriors == [1.0, 2.0, 3.0, 10.0, 5.0]
        assert samples.weights == [1.0, 1.0, 1.0, 1.0, 1.0]
        assert samples.total_samples == 500
        assert samples.log_evidence == 2
        assert samples.unconverged_sample_size == 300
        assert samples.time == 4
        assert samples.number_live_points == 5



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from .aggregator import *
from .model import *


def open_database(
        filename
):
    engine = create_engine(
        f'sqlite:///{filename}'
    )
    session = sessionmaker(bind=engine)()
    Base.metadata.create_all(engine)
    return session



from os import path

import pytest

import autofit as af
from autofit.mock.mock import MockClassx4
from autofit.non_linear.samples import MCMCSamples, Sample
from autofit.non_linear.mcmc.auto_correlations import AutoCorrelations

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


@pytest.fixture(name="samples")
def make_samples():
    model = af.ModelMapper(mock_class_1=MockClassx4)
    print(model.path_priors_tuples)

    parameters = [
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [21.0, 22.0, 23.0, 24.0],
        [0.0, 1.0, 2.0, 3.0],
    ]

    samples = [
        Sample(
            log_likelihood=1.0,
            log_prior=0.0,
            weights=1.0,
            mock_class_1_one=0.0,
            mock_class_1_two=1.0,
            mock_class_1_three=2.0,
            mock_class_1_four=3.0
        ),
        Sample(
            log_likelihood=2.0,
            log_prior=0.0,
            weights=1.0,
            mock_class_1_one=0.0,
            mock_class_1_two=1.0,
            mock_class_1_three=2.0,
            mock_class_1_four=3.0
        ),
        Sample(
            log_likelihood=3.0,
            log_prior=0.0,
            weights=1.0,
            mock_class_1_one=0.0,
            mock_class_1_two=1.0,
            mock_class_1_three=2.0,
            mock_class_1_four=3.0
        ),
        Sample(
            log_likelihood=10.0,
            log_prior=0.0,
            weights=1.0,
            mock_class_1_one=21.0,
            mock_class_1_two=22.0,
            mock_class_1_three=23.0,
            mock_class_1_four=24.0
        ),
        Sample(
            log_likelihood=5.0,
            log_prior=0.0,
            weights=1.0,
            mock_class_1_one=0.0,
            mock_class_1_two=1.0,
            mock_class_1_three=2.0,
            mock_class_1_four=3.0
        )
    ]

    return MCMCSamples(
        model=model,
        samples=samples,
        auto_correlations=AutoCorrelations(
            times=1,
            check_size=2,
            required_length=3,
            change_threshold=4,
            previous_times=5
        ),
        total_walkers=5,
        total_steps=6,
        time=7,
    )


class TestJsonCSV:
    def test__from_csv_table_and_json_info(self, samples):
        mcmc = af.Emcee()
        mcmc.paths = af.DirectoryPaths(path_prefix=path.join("non_linear", "emcee"))

        mcmc.paths._identifier = "tag"

        samples.write_table(filename=path.join(mcmc.paths.samples_path, "samples.csv"))
        samples.info_to_json(filename=path.join(mcmc.paths.samples_path, "info.json"))

        model = af.ModelMapper(mock_class_1=MockClassx4)

        samples = mcmc.samples_via_csv_json_from_model(model=model)

        assert samples.parameters == [
            [0.0, 1.0, 2.0, 3.0],
            [0.0, 1.0, 2.0, 3.0],
            [0.0, 1.0, 2.0, 3.0],
            [21.0, 22.0, 23.0, 24.0],
            [0.0, 1.0, 2.0, 3.0],
        ]
        assert samples.log_likelihoods == [1.0, 2.0, 3.0, 10.0, 5.0]
        assert samples.log_priors == [0.0, 0.0, 0.0, 0.0, 0.0]
        assert samples.log_posteriors == [1.0, 2.0, 3.0, 10.0, 5.0]
        assert samples.weights == [1.0, 1.0, 1.0, 1.0, 1.0]
        assert samples.auto_correlations.times == pytest.approx([31.98507049, 36.51001152, 73.4762926, 67.67495736], 1.0e-4)
        assert samples.auto_correlations.check_size == 2
        assert samples.auto_correlations.required_length == 3
        assert samples.auto_correlations.change_threshold == 4
        assert samples.total_walkers == 5
        assert samples.total_steps == 6
        assert samples.auto_correlations.times[-1] == pytest.approx(67.67, 1.0e-4)



import json
import os
import shutil
from os import path

import dill

from autoconf import conf
from autofit.text import formatter
from .abstract import AbstractPaths, make_path
from ..samples import load_from_table


class DirectoryPaths(AbstractPaths):
    def _path_for_pickle(
            self,
            name: str
    ):
        return path.join(
            self._pickle_path,
            f"{name}.pickle"
        )

    def save_object(
            self,
            name: str,
            obj: object
    ):
        """
        Serialise an object using dill and save it to the pickles
        directory of the search.

        Parameters
        ----------
        name
            The name of the object
        obj
            A serialisable object
        """
        with open(
                self._path_for_pickle(
                    name
                ),
                "wb"
        ) as f:
            dill.dump(
                obj, f
            )

    def load_object(
            self,
            name: str
    ):
        """
        Load a serialised object with the given name.

        e.g. if the name is 'model' then pickles/model.pickle is loaded.

        Parameters
        ----------
        name
            The name of a serialised object

        Returns
        -------
        The deserialised object
        """
        with open(
                self._path_for_pickle(
                    name
                ),
                "rb"
        ) as f:
            return dill.load(
                f
            )

    def remove_object(
            self,
            name: str
    ):
        """
        Remove the object with the given name from the pickles folder.

        Parameters
        ----------
        name
            The name of a pickle file excluding .pickle
        """
        try:
            os.remove(
                self._path_for_pickle(
                    name
                )
            )
        except FileNotFoundError:
            pass

    def is_object(
            self,
            name: str
    ) -> bool:
        """
        Is there a file pickles/{name}.pickle?
        """
        return os.path.exists(
            self._path_for_pickle(
                name
            )
        )

    @property
    def is_complete(self) -> bool:
        """
        Has the search been completed?
        """
        return path.exists(
            self._has_completed_path
        )

    def completed(self):
        """
        Mark the search as complete by saving a file
        """
        open(self._has_completed_path, "w+").close()

    def load_samples(self):
        return load_from_table(
            filename=self._samples_file
        )

    def save_samples(self, samples):
        """
        Save the final-result samples associated with the phase as a pickle
        """
        samples.write_table(filename=self._samples_file)
        samples.info_to_json(filename=self._info_file)

    def load_samples_info(self):
        with open(self._info_file) as infile:
            return json.load(infile)

    def save_all(self, search_config_dict, info, pickle_files):
        self._save_search(config_dict=search_config_dict)
        self._save_model_info(model=self.model)
        self._save_parameter_names_file(model=self.model)
        self.save_object("info", info)
        self.save_object("search", self.search)
        self.save_object("model", self.model)
        self._save_metadata(
            search_name=type(self.search).__name__.lower()
        )
        self._move_pickle_files(pickle_files=pickle_files)

    @property
    @make_path
    def _pickle_path(self) -> str:
        """
        This is private for a reason - use the save_object etc. methods to save and load pickles
        """
        return path.join(self._make_path(), "pickles")

    def _save_metadata(self, search_name):
        """
        Save metadata associated with the phase, such as the name of the pipeline, the
        name of the phase and the name of the dataset being fit
        """
        with open(path.join(self._make_path(), "metadata"), "a") as f:
            f.write(f"""name={self.name}
            non_linear_search={search_name}
            """)

    def _move_pickle_files(self, pickle_files):
        """
        Move extra files a user has input the full path + filename of from the location specified to the
        pickles folder of the Aggregator, so that they can be accessed via the aggregator.
        """
        if pickle_files is not None:
            [shutil.copy(file, self._pickle_path) for file in pickle_files]

    def _save_model_info(self, model):
        """Save the model.info file, which summarizes every parameter and prior."""
        with open(path.join(
                self.output_path,
                "model.info"
        ), "w+") as f:
            f.write(f"Total Free Parameters = {model.prior_count} \n\n")
            f.write(model.info)

    def _save_parameter_names_file(self, model):
        """Create the param_names file listing every parameter's label and Latex tag, which is used for *corner.py*
        visualization.

        The parameter labels are determined using the label.ini and label_format.ini config files."""

        parameter_names = model.model_component_and_parameter_names
        parameter_labels = model.parameter_labels
        subscripts = model.subscripts
        parameter_labels_with_subscript = [f"{label}_{subscript}" for label, subscript in
                                           zip(parameter_labels, subscripts)]

        parameter_name_and_label = []

        for i in range(model.prior_count):
            line = formatter.add_whitespace(
                str0=parameter_names[i], str1=parameter_labels_with_subscript[i], whitespace=70
            )
            parameter_name_and_label += [f"{line}\n"]

        formatter.output_list_of_strings_to_file(
            file=path.join(
                self.samples_path,
                "model.paramnames"
            ),
            list_of_strings=parameter_name_and_label
        )

    @property
    def _info_file(self) -> str:
        return path.join(self.samples_path, "info.json")

    @property
    def _has_completed_path(self) -> str:
        """
        A file indicating that a `NonLinearSearch` has been completed previously
        """
        return path.join(self.output_path, ".completed")

    @make_path
    def _make_path(self) -> str:
        """
        Returns the path to the folder at which the metadata should be saved
        """

        return path.join(
            conf.instance.output_path,
            self.path_prefix,
            self.name,
            self.identifier,
        )



from abc import ABC, abstractmethod
from functools import reduce
from itertools import chain
from operator import and_
from typing import Optional, Tuple, Union, Iterator
from inspect import getfullargspec

import numpy as np

from autofit.mapper.variable import Variable


class AbstractMessage(ABC):
    log_base_measure: float
    _multivariate: bool = False
    _parameter_support: Optional[Tuple[Tuple[float, float], ...]] = None
    _support: Optional[Tuple[Tuple[float, float], ...]] = None

    @property
    @abstractmethod
    def natural_parameters(self):
        pass

    @abstractmethod
    def sample(self, n_samples: Optional[int] = None):
        pass

    @staticmethod
    @abstractmethod
    def invert_natural_parameters(natural_parameters: np.ndarray
    ) -> Tuple[np.ndarray,  ...]:
        pass

    @staticmethod
    @abstractmethod
    def to_canonical_form(x: np.ndarray) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def log_partition(self) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def mean(self) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def variance(self) -> np.ndarray:
        pass

    @property
    def scale(self) -> np.ndarray:
        return self.variance ** 0.5

    def __init__(self, parameters: Tuple[np.ndarray, ...], log_norm=0.):
        self.log_norm = log_norm
        self._broadcast = np.broadcast(*parameters)
        if self.shape:
            self.parameters = tuple(
                np.asanyarray(p) for p in parameters)
        else:
            self.parameters = tuple(parameters)


    def __iter__(self) -> Iterator[np.ndarray]:
        return iter(self.parameters)

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._broadcast.shape

    @property
    def ndim(self) -> int: 
        return self._broadcast.ndim

    @classmethod
    def from_natural_parameters(
            cls, 
            parameters: Tuple[np.ndarray, ...], 
            **kwargs
    ) -> "AbstractMessage":
        args = cls.invert_natural_parameters(parameters)
        return cls(*args, **kwargs)

    @classmethod
    @abstractmethod
    def invert_sufficient_statistics(cls, sufficient_statistics: np.ndarray
    ) -> Tuple[np.ndarray, ...]:
        pass

    @classmethod
    def from_sufficient_statistics(cls, suff_stats: np.ndarray, **kwargs
    ) -> "AbstractMessage":
        natural_params = cls.invert_sufficient_statistics(suff_stats)
        return cls.from_natural_parameters(natural_params, **kwargs)

    def sum_natural_parameters(self, *dists: "AbstractMessage"
    ) -> "AbstractMessage":
        """return the unnormalised result of multiplying the pdf
        of this distribution with another distribution of the same
        type
        """
        new_params = sum(
            (dist.natural_parameters for dist in self._iter_dists(dists) 
            if isinstance(dist, AbstractMessage)),
            self.natural_parameters)
        mul_dist = self.from_natural_parameters(new_params)
        return mul_dist

    def sub_natural_parameters(self, other: "AbstractMessage"
    ) -> "AbstractMessage":
        """return the unnormalised result of dividing the pdf
        of this distribution with another distribution of the same
        type"""
        log_norm = self.log_norm - other.log_norm
        new_params = (self.natural_parameters - other.natural_parameters)
        div_dist = self.from_natural_parameters(new_params, log_norm=log_norm)
        return div_dist

    _multiply = sum_natural_parameters
    _divide = sub_natural_parameters

    def __mul__(self, other: "AbstractMessage") -> "AbstractMessage":
        if np.isscalar(other):
            log_norm = self.log_norm + np.log(other)
            return type(self)(*self.parameters, log_norm=log_norm)
        else:
            return self._multiply(other)

    def __rmul__(self, other: "AbstractMessage") -> "AbstractMessage":
        return self * other

    def __truediv__(self, other: "AbstractMessage") -> "AbstractMessage":
        if np.isscalar(other):
            log_norm = self.log_norm - np.log(other)
            return type(self)(*self.parameters, log_norm=log_norm)
        else:
            return self._divide(other)

    def __pow__(self, other: "AbstractMessage") -> "AbstractMessage":
        natural = self.natural_parameters
        new_params = other * natural
        log_norm = other * self.log_norm
        return self.from_natural_parameters(new_params, log_norm=log_norm)

    def __str__(self) -> str:
        param_attrs = [
            (attr, np.asanyarray(getattr(self, attr)))
            for attr in getfullargspec(self.__init__).args[1:-1]]
        if self.shape:
            pad = max(len(attr) for attr, _ in param_attrs)
            attr_str = "    {:<%d}={}" % pad
            param_strs = ',\n'.join(
                attr_str.format(
                    attr, np.array2string(val, prefix=' '*(pad + 5)))
                for attr, val in param_attrs)
            return f"{type(self).__name__}(\n{param_strs})"
        else:
            param_strs = ', '.join(
                attr + '=' + 
                np.array2string(val, prefix=' '*(len(attr) + 1))
                for attr, val in param_attrs)
            return f"{type(self).__name__}({param_strs})"

    __repr__ = __str__

    def pdf(self, x: np.ndarray) -> np.ndarray:
        return np.exp(self.logpdf(x))

    def logpdf(self, x: np.ndarray) -> np.ndarray:
        shape = np.shape(x)
        if shape:
            x = np.asanyarray(x)

        if shape == self.shape:
            eta = self.natural_parameters
            t = self.to_canonical_form(x)
            log_base = self.log_base_measure
            # TODO this can be made more efficient using tensordot
            eta_t = np.multiply(eta, t).sum(0)  
            return log_base + eta_t - self.log_partition
        elif shape[1:] == self.shape:
            eta = self.natural_parameters
            t = self.to_canonical_form(x)
            eta_t = np.multiply(eta[:, None, ...], t).sum(0)
            return self.log_base_measure + eta_t - self.log_partition

        raise ValueError(
            f"shape of passed value {shape} does not "
            f"match message shape {self.shape}")

    def numerical_logpdf_gradient(self, x: np.ndarray, eps: float=1e-6
    ) -> Tuple[np.ndarray, np.ndarray]:
        shape = np.shape(x)
        if shape:
            x0 = np.array(x, dtype=np.float64)
            logl0 = self.logpdf(x0)
            if self._multivariate:
                grad_logl = np.empty(logl0.shape + x0.shape)
                sl = tuple(slice(None) for _ in range(logl0.ndim))
                with np.nditer(x0, flags=['multi_index'], op_flags=['readwrite']) as it:
                    for xv in it:
                        xv += eps
                        logl = self.logpdf(x0)
                        grad_logl[sl + it.multi_index] = (logl - logl0)/eps
                        xv -= eps
            else:
                l0 = logl0.sum()
                grad_logl = np.empty_like(x0)
                with np.nditer(x0, flags=['multi_index'], op_flags=['readwrite']) as it:
                    for xv in it:
                        xv += eps
                        logl = self.logpdf(x0).sum()
                        grad_logl[it.multi_index] = (logl - l0)/eps
                        xv -= eps
        else:
            logl0 = self.logpdf(x)
            grad_logl = (self.logpdf(x + eps) - logl0)/eps

        return logl0, grad_logl

    def numerical_logpdf_gradient_hessian(self, x: np.ndarray, eps: float=1e-6
    ) -> Tuple[np.ndarray, np.ndarray]:
        shape = np.shape(x)
        if shape:
            x0 = np.array(x, dtype=np.float64)
            if self._multivariate:
                logl0, gradl0 = self.numerical_logpdf_gradient(x0)
                hess_logl = np.empty(gradl0.shape + x0.shape)
                sl = tuple(slice(None) for _ in range(gradl0.ndim))
                with np.nditer(x0, flags=['multi_index'], op_flags=['readwrite']) as it:
                    for xv in it:
                        xv += eps
                        _, gradl = self.numerical_logpdf_gradient(x0)
                        hess_logl[sl + it.multi_index] = (gradl - gradl0)/eps
                        xv -= eps
            else:
                logl0 = self.logpdf(x0)
                l0 = logl0.sum()
                grad_logl = np.empty_like(x0)
                hess_logl = np.empty_like(x0)
                with np.nditer(x0, flags=['multi_index'], op_flags=['readwrite']) as it:
                    for xv in it:
                        xv += eps
                        l1 = self.logpdf(x0).sum()
                        xv -= 2 * eps
                        l2 = self.logpdf(x0).sum()
                        g1 = (l1 - l0)/eps
                        g2 = (l0 - l2)/eps
                        grad_logl[it.multi_index] = g1
                        hess_logl[it.multi_index] = (g1 - g2)/eps
                        xv += eps

                gradl0 = grad_logl
        else:
            logl0 = self.logpdf(x)
            logl1 = self.logpdf(x + eps)
            logl2 = self.logpdf(x - eps)
            gradl0 = (logl1 - logl0)/eps
            gradl1 = (logl0 - logl2)/eps
            hess_logl = (gradl0 - gradl1)/eps

        return logl0, gradl0, hess_logl

    logpdf_gradient = numerical_logpdf_gradient
    logpdf_gradient_hessian = numerical_logpdf_gradient_hessian

    @classmethod
    def project(cls, samples: np.ndarray, log_weights: np.ndarray
    ) -> "AbstractMessage":
        """Calculates the sufficient statistics of a set of samples
        and returns the distribution with the appropriate parameters
        that match the sufficient statistics
        """
        # if weights aren't passed then equally weight all samples

        # Numerically stable weighting for very small/large weights
        log_w_max = np.max(log_weights, axis=0, keepdims=True)
        w = np.exp(log_weights - log_w_max)
        norm = w.mean(0)
        log_norm = np.log(norm) + log_w_max[0]

        tx = cls.to_canonical_form(samples)
        w /= norm
        suff_stats = (tx * w[None, ...]).mean(1)

        assert np.isfinite(suff_stats).all()
        return cls.from_sufficient_statistics(suff_stats, log_norm=log_norm)

    @classmethod
    def from_mode(cls, mode: np.ndarray, covariance: np.ndarray
    ) -> "AbstractMessage":
        pass

    def log_normalisation(self, *dists: Union["AbstractMessage", float]
    ) -> np.ndarray:
        """
        Calculates the log of the integral of the product of a
        set of distributions

        NOTE: ignores log normalisation
        """
        # Remove floats from messages passed
        dists = [
            dist for dist in self._iter_dists(dists) 
            if isinstance(dist, AbstractMessage)]

        # Calculate log product of message normalisation
        log_norm = self.log_base_measure - self.log_partition
        log_norm += sum(
            dist.log_base_measure - dist.log_partition
            for dist in dists)

        # Calculate log normalisation of product of messages
        prod_dist = self.sum_natural_parameters(*dists)
        log_norm -= prod_dist.log_base_measure - prod_dist.log_partition

        return log_norm

    @staticmethod
    def _iter_dists(dists) -> Iterator[Union["AbstractMessage", float]]:
        for elem in dists:
            if isinstance(elem, AbstractMessage):
                yield elem
            elif np.isscalar(elem):
                yield elem
            else:
                for dist in elem:
                    yield dist

    def update_invalid(self, other: 'AbstractMessage') -> 'AbstractMessage':
        valid = self.check_valid()
        if self.ndim:
            valid_parameters = (
                np.where(valid, p, p_safe) for p, p_safe in zip(self, other))
        else:
            # TODO: Fairly certain this would not work
            valid_parameters = self if valid else other  
        return type(self)(*valid_parameters, log_norm=self.log_norm)

    def check_support(self) -> np.ndarray:
        if self._parameter_support is not None:
            return reduce(
                and_,
                ((p >= support[0]) & (p <= support[1])
                 for p, support in 
                 zip(self.parameters, self._parameter_support)))
        elif self.ndim:
            return np.array(True, dtype=bool, ndmin=self.ndim)
        return np.array([True])

    def check_finite(self) -> np.ndarray:
        return np.isfinite(self.natural_parameters).all(0)

    def check_valid(self) -> np.ndarray:
        return self.check_finite() & self.check_support()

    @property
    def is_valid(self) -> bool:
        return np.all(self.check_finite()) and np.all(self.check_support())

    @staticmethod
    def _get_mean_variance(mean: np.ndarray, covariance: float
                           ) -> Tuple[np.ndarray, np.ndarray]:
        mean, covariance = np.asanyarray(mean), np.asanyarray(covariance)

        if not covariance.shape:
            # If variance is float simply pass through
            variance = covariance * np.ones_like(mean)
            if not variance.shape:
                variance = variance.item()
        elif mean.shape == covariance.shape:
            variance = np.asanyarray(covariance)
        elif covariance.shape == mean.shape * 2:
            # If 2D covariance matrix passed get diagonal
            inds = tuple(np.indices(mean.shape))
            variance = np.asanyarray(covariance)[inds * 2]
        else:
            raise ValueError(
                f"shape of covariance {covariance.shape} is invalid "
                f"must be (), {mean.shape}, or {mean.shape * 2}")
        return mean, variance

    def __call__(
        self, 
        x: np.ndarray, 
        _variables: Optional[Tuple[str]]=('x')
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        if _variables is None:
            return self.logpdf(x)
        else:
            if 'x' in _variables:
                loglike, g = self.logpdf_gradient(x)
                g = np.expand_dims(g, list(range(loglike.ndim)))
                return loglike, (g,)
            else:
                return self.logpdf(x), ()

    def as_factor(self, variable: "Variable", name: Optional[str]=None
    ) -> "FactorJacobian":
        from autofit.graphical import FactorJacobian
        if name is None:
            shape = self.shape
            clsname = type(self).__name__
            family = clsname[:-7] if clsname.endswith('Message') else clsname
            name = f"{family}Likelihood" + (str(shape) if shape else '')

        return FactorJacobian(self, x=variable, name=name, vectorised=True)
